import unittest
from pathlib import Path
from pyssian import GaussianInFile

TEST_FILEDIR = Path(__file__).parent.resolve() / 'test_files'

class GaussianInFileTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.testfiles = TEST_FILEDIR.glob('gaussian_input_*.{gjf,com}')
        cls.files_as_text = []
        for gaufile in cls.testfiles: 
            with open(gaufile,'r') as F: 
                cls.files_as_text.append(F.read())

    def test_read_str_equal(self):
        msg = 'Read file does not match its str representation'
        for gaufile,sol in zip(self.testfiles,self.files_as_text):
            with GaussianInFile(gaufile) as GIF: 
                GIF.read()
            test = str(GIF)
            self.assertEqual(test,sol,msg)

    def test_nprocshared(self): 
        GIF = GaussianInFile()
        GIF.nprocs = 24
        msg = 'incorrect nprocshared keyword creation through property nprocs'
        self.assertTrue('nprocshared' in GIF.preprocessing,msg)
        self.assertEqual(24, GIF.preprocessing['nprocshared'],msg)
        GIF.nprocs = 12
        msg = 'incorrect nprocshared keyword modification through property nprocs'
        self.assertTrue('nprocshared' in GIF.preprocessing,msg)
        self.assertEqual(12, GIF.preprocessing['nprocshared'],msg)

        


from pyssian.chemistryutils import is_basis,is_method
import unittest

class ChemistryUtilsTest(unittest.TestCase):
    def setUp(self):
        self.Valid_Basis = '6-311+g(d,p) 6-31g* cc-pVTZ D95V* LanL2DZ SDD28 Def2SVP UGBS2P2++'.split()
        self.Fake_Basis = '6-311g+(d,p) 6-31*g ccpVTZ D96V* LanL2TZ SDD Def2SP UGBS2++P2'.split()
        self.Valid_Methods = 'ub3lyp mp2 casscf ccsd(t) rm062x WB97XD pbepbe'.split()
        self.Fake_Methods = 'bu3lyp pm2 bw97xd m06-2x pbepbe0'.split()
        self.Usual_Keywords = 'opt freq scrf scf #p calcfc empiricaldispersion'.split()
    def test_valid_isbasis(self):
        msg = 'Valid basis not properly recognized'
        for valid in self.Valid_Basis:
            self.assertTrue(is_basis(valid),msg)
    def test_fake_isbasis(self):
        msg = 'Fake basis recognized as valid'
        for fake in self.Fake_Basis:
            self.assertFalse(is_basis(fake),msg)
    def test_valid_ismethod(self):
        msg = 'Valid method not properly recognized'
        for valid in self.Valid_Methods:
            self.assertTrue(is_method(valid),msg)
    def test_fake_ismethod(self):
        msg = 'Fake method recognized as valid'
        for fake in self.Fake_Methods:
            self.assertFalse(is_method(fake),msg)
    def test_usual_keywords(self):
        msg1 = 'Keyword recognized as basis'
        msg2 = 'Keyword recognized as method'
        for keyword in self.Usual_Keywords:
            self.assertFalse(is_basis(keyword),msg1)
            self.assertFalse(is_method(keyword),msg2)



"""
Contains all the hardcoded basis and functionals as well as the neccesary
regex expressions to identify if a certain string is a valid basis or a valid
methods. It also contains a dictionary that maps Atomic Symbol with Atomic
Number and vice versa.
"""

import re

######################## METHODS SECTION #####################################
shellWaveFunctions = 'RO|U|R|'
mutlipleMethodRegex = 'ONIOM|IRCMAX'

methodsRegex = ['DFT','MM','Amber','Dreiding','UFF','AM1','PM3','PM3MM','PM6',
                'PDDG','HF','HFS','XAlpha','HFB','VSXC','HCTH','HCTH93',
                'HCTH147','HCTH407','tHCTH','M06L','B97D','LSDA','LC-wPBE',
                'CAM-B3LYP','wB97XD','wB97','wB97X','LC-BLYP','B3LYP','B3P86',
                'B3PW91','B1B95','mPW1PW91','mPW1LYP','mPW1PBE','mPW3PBE',
                'B98','B971','B972','PBE1PBE','B1LYP','O3LYP','TPSSh','BMK',
                'M062X','M06','M06HF','M05','M052X','X3LYP','BHandH',
                'BHandHLYP','tHCTHhyb','HSEh1PBE','HSE2PBE','HSEhPBE',
                'PBEh1PBE','CASSCF','CAS','MP2','MP3','MP4','MP5','B2PLYP',
                'B2PLYPD','mPW2PLYP','mPW2PLYPD','QCISD','CCD',r'CCSD\(T\)',
                'CCSD','CC','QCID','BD','EPT','CBS-4M','CBS-QB3','ROCBS-QB3',
                'CBS-APNO','G1','G2','G2MP2','G3','G3MP2','G3B3','G3MP2B3','G4',
                'G4MP2','W1U','W1BD','W1RO','CIS',r'CIS\(D\)','CID','CISD','TD',
                'EOMCCSD','ZINDO','DFTB','DFTBA','GVB','CNDO','INDO','MINDO',
                'MNDO','NMR','SAC-CI']
methodsRegex = [i.upper() for i in methodsRegex]
methodsRegex = '|'.join(methodsRegex)

exchangeFunctional = ['S','XA','B','PW91','MPW','G96','PBE','OPBE','O','TPSS',
                      'BRX','PKZB','WPBEH','PBEH','LC-']
exchangeFunctional = '|'.join(exchangeFunctional)

correlationFunctional = ['VWN','VWN5','LYP','PL','P86','PW91','B95','PBE',
                         'TPSS','KCIS','BRC','PKZB','VP86','V5LYP']
correlationFunctional = '|'.join(correlationFunctional)

methodRegex = f'({shellWaveFunctions})({methodsRegex})'
compoundMethodRegex = '({0})({1})[0-9]*({2})'.format(shellWaveFunctions,
                                                 exchangeFunctional,
                                                 correlationFunctional)

method_re = re.compile(methodRegex)
multiplemethod_re = re.compile(mutlipleMethodRegex)
compoundmethod_re = re.compile(compoundMethodRegex)


###################### THE PRETTY FUNCTION ######################
def is_method(candidate):
    """
    Tests if a candidate string is a valid method recognized by Gaussian.

    Parameters
    ----------
    candidate : str

    Returns
    -------
    bool

    """
    candidate = candidate.upper()

    test1 = multiplemethod_re.match(candidate)
    test2 = method_re.match(candidate)
    test3 = compoundmethod_re.match(candidate)
    if test1 is not None:
        test1 = test1[0] == candidate
    if test2 is not None:
        test2 = test2[0] == candidate
    if test3 is not None:
        test3 = test3[0] == candidate
    return test1 or test2 or test3

######################## BASIS SECTION #########################################
## Pople Bases require lowercasing for design reasons.
Pople = dict(primitives = r'sto|mc|[0-9]',
             zeta = '[0-9]{1,3}',
             diffuse = r'[\+]{0,2}',
             pol1 = r'[0-9]?(d|f|d\'|f\'){1,2}',
             pol2 = r'[0-9]?(\,p|d|p\'|d\')?')
x1 = Pople['pol1']
x2 = Pople['pol2']
Pople['polarization'] = r'\(({pol1}{pol2})\)|\*|\*\*'.format(pol1=x1,pol2=x2)
Pople_Basis = r'({primitives})\-({zeta})({diffuse})g({polarization})?'
Pople_Basis = Pople_Basis.format(**Pople)
# The resulting regex is this one but I wanted the reasoning to be written down
# (sto|mc|[0-9])\-([0-9]{1,3})([\+]{0,2})g(\([0-9]{0,1}(d|f|d\'|f\'){1,2}\,{0,1}[0-9]{0,1}(p|d|p\'|d\'){0,1}\)){0,1}

## ccBases
## Assumes lowercased, it's simpler so its not divided
cc_Basis = r'((sp|d)?t?(aug|jul|jun|may|apr){1}\-)?cc\-pv(d|t|q|5|6)z'
#((sp|d){0,1}t{0,1}(aug|jul|jun|may|apr){1}\-){0,1}cc\-pv(d|t|q|5|6)z

# UGBS
UGBS_Basis = r'ugbs[1-9](p|v|o)(2?\+{1,2})?'

# SV Family
SV_Basis = r'(def2)?(sv|tzv?|qzv?)p{0,2}' #Requires the usage of match

# Family that includes indicating the number of core electrons
core_Basis = r'(sdd|shf|sdf|mhf|mdf|mwb|oldsdd|sddall)[0-9]{1,3}'

# D95 Family
D95_Basis = 'd95v?('+Pople['polarization']+')'

# CEP family
CEP_Basis = r'cep\-?[0-9]{1,3}g('+Pople['polarization']+')'

# SHC Basis
SHC_Basis = 'shc('+Pople['polarization']+')'

# Or to hardcode all the family

Hardcoded_basis = set(['sec','lanl2mb','lanl2dz','midix','epr-ii','epr-iii',
                 'mtsmall','dgdzvp','dgdzvp2','dgtzvp','cbsb7','gen','genecp'])

# Construct the full regex
all_basis = [Pople_Basis,UGBS_Basis,cc_Basis,SV_Basis,core_Basis,
             D95_Basis,CEP_Basis,SHC_Basis]

basis_expr = '|'.join([f'({i})' for i in all_basis])
#print(basis_expr)
basis_regex = re.compile(basis_expr)
######################## THE PRETTY FUNCTION ###################################
def is_basis(candidate):
    """
    Tests if a candidate string is a valid basis set recognized by Gaussian.

    Parameters
    ----------
    candidate : str

    Returns
    -------
    bool

    """
    lowercandidate = candidate.lower()
    in_hardcoded = lowercandidate in Hardcoded_basis
    match = basis_regex.match(lowercandidate) # Has to be match and not search
    if match:
        in_regex = match[0] == lowercandidate
    else:
        in_regex = False
    return  in_hardcoded or in_regex
######################### PERIODIC TABLE #######################################
# It is important that X appears the first one (enumeration starts in 0)
# here X stands for a dummy atom
items = """X
     H                                                                                                  He
    Li Be  B                                                                             C   N   O   F  Ne
    Na Mg Al                                                                            Si   P   S  Cl  Ar
     K Ca Sc                                           Ti  V Cr Mn Fe Co Ni Cu  Zn  Ga  Ge  As  Se  Br  Kr
    Rb Sr  Y                                           Zr Nb Mo Tc Ru Rh Pd Ag  Cd  In  Sn  Sb  Te   I  Xe
    Cs Ba La Ce Pr Nd Pm Sm Eu Gd Tb Dy Ho Er Tm Yb Lu Hf Ta  W Re Os Ir Pt Au  Hg  Tl  Pb  Bi  Po  At  Rn
    Fr Ra Ac Th Pa  U Np Pu Am Cm Bk Cf Es Fm Md No Lr Rf Db Sg Bh Hs Mt Ds Rg Uub Uut Uuq Uup Uuh Uus Uuo
    """
items = items.replace('\n',' ').strip().split()
PeriodicTable = dict()
for i,Sym in enumerate(items):
    PeriodicTable[i] = Sym        #  1  -> 'H'
    PeriodicTable[str(i)] = Sym   # '1' -> 'H'
    PeriodicTable[Sym] = i        # 'H' ->  1



"""
Contains auxiliary classes, usefull in combination with the gaussianclasses.
"""
import os
from math import ceil
from itertools import cycle
from pathlib import Path

import numpy

from .chemistryutils import PeriodicTable

class Geometry(object):
    """
    This class provides a basic interface and different construction methods 
    to simplify the extraction of the geometry of a molecule from different 
    sources 
    
    Currently: Link202, GaussianInFile, xyz file
    """
    # classattribute for object naming
    counter = 1
    def __init__(self):
        """ Initialize an empty geometry with generic title """
        self.coordinates = []
        self.atoms = []
        cls = type(self)
        self.title = str(cls.counter)
        cls.counter += 1
    def __repr__(self):
        return f'< Geometry {self.title} object >'
    def __str__(self):
        linef = " {:<}\t{: 0.09f}\t{: 0.09f}\t{: 0.09f}"
        As = self.atoms
        XYZ = self.coordinates
        text = "\n".join([linef.format(str(atom),*xyz)
                                        for atom,xyz in zip(As,XYZ)])
        return text
    def write(self,File):
        """ Write Geometry to a File """
        File.write(str(self))

    def update_atoms(self,other):
        """ Updates the atoms names from a mapping """
        self.atoms = [other[atom] for atom in self.atoms]

    @classmethod
    def from_L202(cls,L202):
        """Generate a Geometry instance from a Gaussian file. If it can't find a
        mapping from atomic Number to symbol it will store the string
        representation of the number """
        Geom = cls()
        Coords = []
        Atoms = []
        orientation = L202.orientation
        for atom in orientation:
            Coords.append(atom[3:])
            Atoms.append(PeriodicTable[atom[1]])
        Geom.atoms = Atoms
        Geom.coordinates = Coords
        return Geom
    @classmethod
    def from_Input(cls,InputFile):
        """Generate a Geometry instance from a Gaussian Input File instance."""
        Geom = cls()
        Coords = []
        Atoms = []
        for line in InputFile.geometry.split('\n'):
            Aux = line.strip().split()
            Atoms.append(Aux[0])
            B = tuple(map(float,Aux[1:]))
            Coords.append(B)
        Geom.atoms = Atoms
        Geom.coordinates = Coords
        return Geom
    @classmethod
    def from_xyz(cls,xyzfile):
        """Generate a Geometry instance from a xyz file."""
        Geom = cls()
        Coords = []
        Atoms = []
        with open(xyzfile,'r') as F:
            _iter = F.__iter__()
            n = int(next(_iter).strip())
            _ = next(_iter)
            for i in range(n):
                line = next(_iter)
                aux = line.strip().split()
                Atoms.append(aux[0])
                B = tuple(map(float,aux[1:]))
                Coords.append(B)
        Geom.atoms = Atoms
        Geom.coordinates = Coords
        return Geom

    def to_xyz(self,title=None):
        if title is None: 
            title = f'{self.title}'
        return f'{len(self.atoms)}\n{title}\n{self}\n'

class Cube(object):
    """
    Representation of a Gaussian .cube File generated with the cubegen tool.
    This class facilitates some operations of the cubeman tool allowing
    operations of more than 2 .cube files.

    Files with more than 1 cube are accepted but correct behaviour is not 
    guaranteed.

    Attributes
    ----------
    natoms : int
        total number of atoms
    type : str
        Type of cube. Currently only {'MO','other'} are considered
    isUnrestricted : bool
        True if the line that contains the number of atoms has 5 elements
        instead of 4.
    origin : tuple
        origin of the cube.
    shape : tuple
        Number of points in each of the three directions.
    basis : list
        Vectors of the three main directions.
    atoms : list
        List of the AtNumber, Charge, (x,y,z) per each atom.
    nMO : int
        Number of molecular orbitals
        (Current implementation is designed for only one MO per Cube)
    MO_Ids : list
        Ids of the MO orbitals.
        (Current implementation is designed for only one MO per Cube)
    matrix : numpy.array
        Matrix of values of the Cube.
    """

    def __init__(self):
        self.natoms = 0
        self.type = 'empty'
        self.isUnrestricted = False
        self.origin = (0,0,0)
        self.shape = (0,0,0)
        self.basis = [(0,0,0),
                      (0,0,0),
                      (0,0,0)]
        self.atoms = []
        self.nMO = 0
        self.MO_Ids = []
        self.matrix = []
    def __repr__(self):
        return f'<{type(self).__name__}_{id(self)}>'

    def __add__(self,other):
        New = self.__class__()
        for attr in 'natoms type origin shape basis atoms nMO'.split():
            if getattr(self,attr) != getattr(self,attr):
                raise ValueError(f'{self}.{attr} != {other}.{attr}')
            else:
                setattr(New,attr,getattr(self,attr))
            New.matrix = self.matrix + other.matrix
        return New
    def __radd__(self,other):
        return self + other
    def __mul__(self,other):
        New = self.__class__()
        for attr in 'natoms type origin shape basis atoms nMO'.split():
            setattr(New,attr,getattr(self,attr))
        New.matrix = self.matrix*other
        return New
    def __rmul__(self,other):
        return self*other
    def __neg__(self):
        return self*(-1)
    def __sub__(self,other):
        return self + (-other)
    def __pow__(self,other):
        New = self.__class__()
        for attr in 'natoms type origin shape basis atoms nMO'.split():
            setattr(New,attr,getattr(self,attr))
        New.matrix = self.matrix**other
        return New

    # Writing functions
    def write(self,OFile):
        """
        Writes a cube to a File. 

        Parameters
        ----------
        OFile : str
            A valid path to a file. 
        """
        # Formats
        atom_format  = '{0: >5}   {1: .6f}   {2: .6f}   {3: .6f}   {4: .6f}\n'.format
        basis_format = '{0: >5}   {1: .6f}   {2: .6f}   {3: .6f}\n'.format

        # preparation
        natoms = str(self.natoms)
        sign = ' '
        if self.type == 'MO':
            sign = '-'
        x0,y0,z0 = self.origin
        (ix,jx,kx), (iy,jy,ky), (iz,jz,kz) =  self.basis
        Nx,Ny,Nz = self.shape
        MO_Ids = self.MO_Ids
        if self.nMO > 0 and not self.MO_Ids:
            MO_Ids = [1,] 
        n_MO_line = ['{: >5}'.format(str(self.nMO)),]
        n_MO_line.extend(list(map('{: >5}'.format, MO_Ids)))
        # Writing
        with open(OFile,'w') as F:
            F.write('Default Title Line\n')
            F.write('Default Other Title Line\n')
            line = f'{sign+natoms: >5}   {x0: .6f}   {y0: .6f}   {z0: .6f}\n'
            if self.isUnrestricted:
                line += '   1'
            F.write(line)
            F.write(basis_format(Nx,ix,jx,kx))
            F.write(basis_format(Ny,iy,jy,ky))
            F.write(basis_format(Nz,iz,jz,kz))
            for atom in self.atoms:
                F.write(atom_format(*atom))
            F.write(''.join(n_MO_line)+'\n')
            self.write_values(F)
    def write_values(self,File):
        """
        Handles the writing in of the values of self.matrix with the fortran77
        format

        Parameters
        ----------
        File : _io.TextIOWrapper
            the output of a open(Filename,'w')
        """
        # Formats
        val_format = ' {: .5E}'.format
        # preparation
        Nx,Ny,Nz = self.shape
        MO_Ids = self.MO_Ids
        counter = 0
        for i in range(Nx):
            for j in range(Ny):
                if len(MO_Ids) > 1:
                    for Id,_ in enumerate(MO_Ids):
                        for k in range(Nz):
                            counter += 1
                            File.write(val_format(self.matrix[i][j][k][Id]))
                            if counter == 6:
                                File.write('\n')
                                counter = 0
                    if counter != 0:
                        counter = 0
                        File.write('\n')
                else:
                    for k in range(Nz):
                        counter += 1
                        File.write(val_format(self.matrix[i][j][k]))
                        if counter == 6:
                            File.write('\n')
                            counter = 0
                    if counter != 0:
                        counter = 0
                        File.write('\n')

    # Constructor methos
    @classmethod
    def from_file(cls,file):
        """
        Creates a Cube instance from a cube file.

        Parameters
        ----------
        file : str
            A valid filename to an existing .cube file.

        Returns
        -------
        Cube
            Instantiated cube from the file.
        """
        newcube = cls()
        with open(file,'r') as F:
            lines = F.readlines()
        _iter = lines.__iter__()
        newcube.handle_title_lines(_iter)
        newcube.read_origin_line(_iter)
        newcube.read_basis_lines(_iter)
        newcube.read_atoms_lines(_iter)
        newcube.read_MO_lines(_iter)
        newcube.read_matrix_lines(_iter)

        return newcube

    # Parsing functions
    def handle_title_lines(self,iterable):
        # Discard first two lines
        _ = next(iterable)    # {Title1} MO={1} || {Title2} MO={2}
        _ = next(iterable)    # MO coeffic + MO coeffic
    def read_origin_line(self,iterable):
        # Read origin and number of atoms
        line = next(iterable) # -{Natoms} {xorigin} {yorigin} {zorigin}
        Aux = line.strip().split()
        natoms,x0,y0,z0 = Aux[:4]
        if len(Aux) > 4:
            self.isUnrestricted = '1' == Aux[5]
        x0,y0,z0 = map(float,(x0,y0,z0))
        if int(natoms) < 0:
            self.type = 'MO'
        else:
            self.type = 'other'
        natoms = abs(int(natoms))
        self.natoms = natoms
        self.origin = (x0,y0,z0)
    def read_basis_lines(self,iterable):
        # Read number of points and directions of the basis
        line = next(iterable) # {Nx} {ix} {jx} {kx}
        Nx, ix,jx,kx = line.strip().split()
        ix,jx,kx = map(float,(ix,jx,kx))
        line = next(iterable) # {Ny} {iy} {jy} {ky}
        Ny, iy,jy,ky = line.strip().split()
        iy,jy,ky = map(float,(iy,jy,ky))
        line = next(iterable) # {Nz} {iz} {jz} {kz}
        Nz, iz,jz,kz = line.strip().split()
        iz,jz,kz = map(float,(iz,jz,kz))
        Nx,Ny,Nz = map(int,(Nx,Ny,Nz))
        self.shape = (Nx,Ny,Nz)
        self.basis = [(ix,jx,kx),(iy,jy,ky),(iz,jz,kz)]
    def read_atoms_lines(self,iterable):
        # Read the atoms, charges, and coordinates
        atoms = []
        for _ in range(self.natoms):
            line = next(iterable)
            AtN,Chrg,x,y,z = line.strip().split()
            AtN = int(AtN)
            Chrg,x,y,z = map(float,(Chrg,x,y,z))
            atoms.append((AtN,Chrg,x,y,z))
        self.atoms = atoms
    def read_MO_lines(self,iterable):
        ## For now I'm gonna treat it as a single line
        # Now I guess this is the "nMO" and "Ids..." line/s
        line = next(iterable)
        nMO,*MO_Ids = line.strip().split()
        self.MO_Ids = MO_Ids
        self.nMO = int(nMO)
    def read_matrix_lines(self,iterable):
        Nx,Ny,Nz = self.shape
        nMO = self.nMO
        nEntries = Nx*Ny
        nVals = Nz*nMO
        nlines = ceil(nVals/6)
        if nMO > 1:
            matrix = numpy.zeros(shape=(Nx,Ny,Nz,nMO))
        else:
            matrix = numpy.zeros(shape=(Nx,Ny,Nz))

        data = []
        values = []
        for index,line in zip(cycle(range(nlines)),iterable):
            values.extend(list(map(numpy.float64,line.strip().split())))
            if index == nlines-1:
                data.append(values)
                values = []
        else:
            if values:
                raise RuntimeError('Wrong developer assumptions on the .cube format')

        for i in range(Nx):
            for j in range(Ny):
                index1 = i*Ny + j
                items = data[index1]
                if nMO > 1:
                    for l in range(nMO):
                        for k in range(Nz):
                            index2 = l*Nz + k
                            matrix[i,j,k,l] = items[index2]
                else:
                    for k in range(Nz):
                        matrix[i,j,k] = items[k]
        self.matrix = matrix

class DirectoryTree(object):
    """
    Class that provides recursive file iteration.
    """
    def __init__(self,path):
        self.root = Path(path)
        self.cwd = Path(os.getcwd())
        self.newroot = self.root
    def set_newroot(self,newroot):
        self.newroot = Path(newroot)

    def newpath(self,path):
        """
        Takes a path rooted to the previous root and prepares it to be
        rooted in the new root.
        """
        # Maybe the usage of the cwd might be necessary
        # so it is usefull this behaviour encapsulated in a function
        return self.newroot.joinpath(path.relative_to(self.root))

    @staticmethod
    def RecursiveFileSystemGenerator(path,key=None):
        """
        Generator that traverses the filesystem in depth first order and yields
        the paths that satisfy the key function condition. If no key function
        is provided, returns all folders and items. The first yield is always
        None, and the second yield is always the path value if it satisfies the
        key condition.

        Parameters
        ----------
        path : Path
            Path pointing to a directory.
        key : function
            A function that returns a boolean. Used in a similar function as
            the 'sorted' python builtin.
        """
        cls = DirectoryTree
        if key is None:
            key=lambda x: True
        yield None
        if key(path):
            yield path
        if path.is_dir():
            for subpath in path.iterdir():
                subgen =cls.RecursiveFileSystemGenerator(subpath,key=key)
                _ = next(subgen)
                for i in subgen:
                    yield i

    @property
    def folders(self):
        generator = self.RecursiveFileSystemGenerator
        _iter = generator(self.root,key=lambda x: x.is_dir())
        _  = next(_iter)
        return _iter

    @property
    def infiles(self):
        def f_in(x):
            Out = False
            if x.exists() and x.suffix == '.in':
                Out = True
            return Out
        _iter = self.RecursiveFileSystemGenerator(self.root,key=f_in)
        _  = next(_iter)
        return _iter

    @property
    def outfiles(self):
        def f_out(x):
            Out = False
            if x.exists() and x.suffix == '.out':
                Out = True
            return Out
        _iter = self.RecursiveFileSystemGenerator(self.root,key=f_out)
        _  = next(_iter)
        return _iter

    def create_folders(self):
        for folder in self.folders:
            newpath = self.newpath(folder)
            newpath.mkdir(parents=True,exist_ok=True)



import setuptools

__version__ = '1.1.0'

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'pyssian',
  version = __version__,
  description = 'Parser Library for Gaussian Files',
  long_description=long_description,
  long_description_content_type="text/x-rst",
  url = 'https://github.com/maserasgroup-repo/pyssian',
  author = 'Raúl Pérez-Soto',
  author_email = 'rperezsoto.research@gmail.com',
  classifiers = ['License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 ],
  keywords = ['compchem, gaussian, parser'],
  packages = setuptools.find_packages(),
  python_requires='>=3.6',
  install_requires=['setuptools','pathlib','numpy'],
  include_package_data=True,
  package_data = {'test_files': ['pyssian/tests/test_files/*.txt'],
                  'tests' : ['pyssian/tests/*.py']},
  project_urls={'Bug Reports': 'https://github.com/maserasgroup-repo/pyssian/issues',
                'Source': 'https://github.com/maserasgroup-repo/pyssian',
                'Docs' : 'https://maserasgroup-repo.github.io/pyssian/'
               },
)



import unittest
from pathlib import Path
from itertools import chain

from pyssian.linkjobparsers import *

TEST_FILEDIR = Path(__file__).parent.resolve() / 'test_files'

SMARK = '\n## Split Here ##\n'
MMARK = '\n## Match ##\n'

class TestGeneralLinkJob(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('General.txt')
        cls.refile = TEST_FILEDIR.joinpath('General_regex.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [GeneralLinkJob(i) for i in txt.split(SMARK)]

    def test_regex(self):
        msg = 're_enter regex does not match properly'
        regex = GeneralLinkJob.re_enter
        with open(self.refile,'r') as F:
            txt = F.read()
        items = [i.strip() for i in txt.split(SMARK)]
        for obj,solution in zip(self.objects,items):
            txt = obj.text
            test = regex.findall(txt)[0]
            self.assertTrue(test == solution,msg)

    def test_init(self):
        msg = 'Incorrect parsing of GeneralLinkJob'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of GeneralLinkJob'
        obj = GeneralLinkJob('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertTrue(obj.number == -1,msg)

class TestLink1(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l1.txt')
        cls.refile = TEST_FILEDIR.joinpath('l1_regex.txt')
        cls.propfile = TEST_FILEDIR.joinpath('l1_properties.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link1(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link1'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,1)
            self.assertTrue(bool(obj.info),msg)

    def test_regex_link0(self):
        msg = 're_link0 regex does not match properly'
        regex = Link1.re_link0
        with open(self.refile) as F:
            txt = F.read()
        items = [i for i in txt.split(SMARK)]
        solutions = [['%nprocshared=8','%mem=18000MB'],[]]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            self.assertTrue(test == solution,msg)

    def test_regex_commandline(self):
        msg = 're_commandline regex does not match properly'
        regex = Link1.re_commandline
        with open(self.refile) as F:
            txt = F.read()
        items = [i for i in txt.split(SMARK)]
        solutions = [['#p opt=(calcfc,ts,noeigentest) freq b3lyp/6-31+g(d) scrf=(solvent=dich\nloromethane,smd) nosymm empiricaldispersion=gd3\n'],
                     ['#P Geom=AllCheck Guess=TCheck SCRF=Check GenChk RB3LYP/6-31+G(d) Freq\n']]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            self.assertTrue(test == solution,msg)

    def test_regex_IOps(self):
        msg = 're_IOps regex does not match properly'
        regex = Link1.re_IOps
        with open(self.refile) as F:
            txt = F.read()
        items = [i for i in txt.split(SMARK)]
        solutions = [['1/5=1,10=4,11=1,14=-1,18=20,26=3,38=1/1,3;',
                      '2/9=110,12=2,15=1,17=6,18=5,40=1/2;',
                      '4//1;',
                      '7/10=1,18=20,25=1,30=1/1,2,3,16;',
                      '1/5=1,11=1,14=-1,18=20,26=3/3(-5);'],
                     ['4/5=101/1;',
                      '5/5=2,53=9,98=1/2;',
                      '8/6=4,10=90,11=11/1;',
                      '99//99;']]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            self.assertTrue(test == solution,msg)

    def test_regex_internaljob(self):
        msg = 're_internaljob regex does not match properly'
        regex = Link1.re_internaljob
        with open(self.refile) as F:
            txt = F.read()
        items = [i for i in txt.split(SMARK)]
        solutions = [[],['2',]]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            self.assertTrue(test == solution,msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link1'
        text = 'Link1:  Proceeding to internal job step number 13.'
        obj = Link1(text,asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,1)
        self.assertTrue(bool(obj.info),msg)

    def test_guess_type(self):
        msg = 'Incorrect type guess'
        text = 'Link1:  Proceeding to internal job step number 13.'
        obj = Link1(text,asEmpty=True)
        types = {0:'Constrained Optimization',
                 1:'Optimization',
                 2:'Frequency Calculation',
                 3:'Unidentified'}
        keywords = ['modredundant','opt','modredundant',
                    'opt','freq','opt','freq',
                    'optimization','#p','redundant']
        solutions = [0,0,0,1,1,1,2,3,3,3]
        while keywords:
            keywords.pop(0)
            obj.commandline = ' '.join(keywords)
            sol = types[solutions.pop(0)]
            test = obj._guess_type()
        self.assertTrue(test == sol,msg)

    def test_commandline(self):
        msg = 'commandline does not maintain its integrity'
        with open(self.propfile) as F:
            txt = F.read()
        items = [i.rstrip('\n').split('\n') for i in txt.split(SMARK)]
        for item,obj in zip(items,self.objects):
            _ , sol = item
            self.assertTrue(obj.commandline == sol,msg)

    def test_info(self):
        msg = 'InternalJobInfo not properly added'
        InternalJobInfo = Link1.InternalJobInfo
        with open(self.propfile) as F:
            txt = F.read()
        items = [i.rstrip('\n').split('\n') for i in txt.split(SMARK)]
        for item,obj in zip(items,self.objects):
            Aux = item[0].split(',')
            sol = InternalJobInfo(int(Aux[0]),Aux[1],Aux[2]=='1')
            self.assertTrue(obj.info == sol,msg)

class TestLink101(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l101.txt')
        cls.charge_spin_solutions = [(-1,2),(0,2),(2,2),(2,2)]
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link101(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link101'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,101)
            self.assertTrue(obj.charge is not None,msg)
            self.assertTrue(obj.spin is not None,msg)

    def test_regex_charge(self):
        msg = 're_charge regex does not match properly'
        regex = Link101.re_charge
        solutions = [str(i) for i,_ in self.charge_spin_solutions]
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test == solution,msg)

    def test_regex_spin(self):
        msg = 're_spin regex does not match properly'
        regex = Link101.re_spin
        solutions = [str(i) for _,i in self.charge_spin_solutions]
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test == solution,msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link101'
        obj = Link101('-> Input Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,101)

    def test_charge_spin(self):
        msg = '{} is not properly parsed'.format
        for item,obj in zip(self.charge_spin_solutions,self.objects):
            charge, spin = item
            self.assertTrue(obj.charge == charge ,msg('charge'))
            self.assertTrue(obj.spin == spin, msg('spin'))

class TestLink103(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l103.txt')
        cls.parametersfile = TEST_FILEDIR.joinpath('l103_parameters.txt')
        cls.derivativesfile = TEST_FILEDIR.joinpath('l103_derivatives.txt')
        cls.convergencefile = TEST_FILEDIR.joinpath('l103_convergence.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link103(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link103'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,103,msg)
            self.assertTrue(bool(obj.mode),msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link101'
        obj = Link103('-> Input Text',asEmpty=True)
        attrs = ['mode','state','convergence', 'parameters', 'derivatives',
                 'stepnumber', 'scanpoint']
        for attr in attrs:
            test = getattr(obj,attr)
            self.assertTrue(not bool(test) or test is None)

    def test_locate_mode(self):
        msg = 'Mode {} not properly recognized'.format
        solutions = ['Init','Iteration','End','End']
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            with self.subTest(Test_Object=i,mode=solution):
                self.assertTrue(obj.mode == solution,msg(solution))

    def test_regex_parameters(self):
        msg = 're_parameters regex does not match properly'
        regex = Link103.re_parameters
        with open(self.parametersfile,'r') as F:
            solutions = [sample.split('\n') for sample in F.read().split(SMARK)]
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            test = regex.findall(obj.text)
            if not solution:
                with self.subTest(object=i,parameter='None'):
                    self.assertFalse(bool(test))
            for j,(t,s) in enumerate(zip(test,solution)):
                with self.subTest(object=i,parameter=j):
                    self.assertEqual(t,s,msg)

    def test_regex_derivatives(self):
        msg = 're_derivatives regex does not match properly'
        regex = Link103.re_derivatives
        solutions = []
        with open(self.derivativesfile,'r') as F:
            txt = F.read()
        for sample in txt.split(SMARK):
            Aux = []
            if sample:
                for s in sample.split('\n'):
                    Aux.append(tuple(s.split(',')))
                solutions.append(Aux)
            else:
                solutions.append([])
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            test = regex.findall(obj.text)
            if not solution:
                with self.subTest(object=i,parameter='None'):
                    self.assertFalse(bool(test))
            for j,(t,s) in enumerate(zip(test,solution)):
                with self.subTest(object=i,parameter=j):
                    self.assertEqual(t,s,msg)

    def test_regex_convergence(self):
        msg = 're_convergence regex does not match properly'
        regex = Link103.re_convergence
        solutions = []
        with open(self.convergencefile,'r') as F:
            txt = F.read()
        for sample in txt.split(SMARK):
            Aux = []
            if sample:
                for s in sample.split('\n'):
                    Aux.append(tuple(s.split(',')))
                solutions.append(Aux)
            else:
                solutions.append([])
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            test = regex.findall(obj.text)
            if not solution:
                with self.subTest(object=i,parameter='None'):
                    self.assertFalse(bool(test))
            for j,(t,s) in enumerate(zip(test,solution)):
                with self.subTest(object=i,parameter=j):
                    self.assertEqual(t,s,msg)

    def test_locate_numbers(self):
        msg = 'stepnumber {} not properly recognized'.format
        solutions = [0,1,420,66]
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            with self.subTest(Test_Object=i,stepnumber=solution):
                self.assertEqual(obj.stepnumber,solution,msg(solution))
        msg = 'scanpoint {} not properly recognized'.format
        solutions = [None,None,None,11]
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            with self.subTest(Test_Object=i,scanpoint=solution):
                self.assertEqual(obj.scanpoint,solution,msg(solution))

class TestLink120(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l120.txt')
        cls.re_energy_solutions = ['',
                                   '',
                                   '',
                                   '-2572.457446471110']
        cls.re_energy_partition_solutions = [[],
                                             [],
                                             [],
                                             [['1', 'low','model','0.403686133797'],
                                              ['2','high','model','-2573.050183259390'],
                                              ['3', 'low', 'real','0.996422922077']]]
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link120(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link120'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,120,msg)
            self.assertTrue(obj.energy is None or bool(obj.energy),msg)

    def test_regex_energy(self):
        msg = 're_energy regex does not match properly'
        regex = Link120.re_energy
        solutions = self.re_energy_solutions
        for obj,solution in zip(self.objects,solutions):
            match = regex.findall(obj.text)
            self.assertEqual(bool(match),bool(solution),msg)
            if match: 
                self.assertEqual(match[0],solution,msg)
    
    def test_regex_energy_partitions(self):
        msg = 're_energy regex does not match properly'
        regex = Link120.re_energy_partitions
        solutions = self.re_energy_partition_solutions
        for obj,solution in zip(self.objects,solutions):
            match = regex.findall(obj.text)
            self.assertEqual(bool(match),bool(solution),msg)
            if match:
                for (p,lev,model,energy),sol in zip(match,solution):
                    self.assertEqual([p,lev,model,energy],sol) 
            
    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link120'
        obj = Link120('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,120)

    def test_energy(self):
        msg = 'Energy value not properly read'
        solutions = self.re_energy_solutions
        for obj,solution in zip(self.objects,solutions):
            test = obj.energy
            self.assertEqual(bool(test),bool(solution),msg)
            if solution: 
                self.assertEqual(test,float(solution),msg)
    
    def test_energy_partition(self):
        msg = 'Energy partitions not properly read'
        solutions = self.re_energy_partition_solutions
        EnergyPartition = Link120._EnergyPartition
        for obj,solution in zip(self.objects,solutions):
            match = obj.energy_partitions
            self.assertEqual(bool(match),bool(solution),msg)
            if not match:
                continue
            for (p,level,model,energy),sol in zip(match,solution):
                p,l,m,e = sol
                sol = EnergyPartition(int(p),l,m,float(e))
                test = EnergyPartition(int(p),level,model,float(energy))
                self.assertEqual(test,sol,msg)
    
class TestLink123(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l123.txt')
        cls.orientationfile = TEST_FILEDIR.joinpath('l123_orientation.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link123(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link123'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,123,msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link101'
        obj = Link123('-> Input Text',asEmpty=True)
        attrs = ['orientation', 'direction', 'step', 'reactioncoord']
        for attr in attrs:
            test = getattr(obj,attr)
            self.assertTrue(not bool(test) or test is None)

    def test_regex_orientation(self):
        msg = 're_orientation regex does not match properly'
        regex = Link123.re_orientation
        solutions = []
        with open(self.orientationfile,'r') as F:
            txt = F.read()
        for sample in txt.split(SMARK):
            Aux = []
            if sample:
                for s in sample.split('\n'):
                    Aux.append(tuple(s.split('\t')))
                solutions.append(Aux)
            else:
                solutions.append([])
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            test = regex.findall(obj.text)
            if not solution:
                with self.subTest(object=i,atom='None'):
                    self.assertFalse(bool(test))
            for j,(t,s) in enumerate(zip(test,solution)):
                with self.subTest(object=i,atom=j):
                    self.assertEqual(t,s,msg)

    def test_locate_irc_step(self):
        msg0 = 'reactioncoord {} not properly recognized'.format
        msg1 = 'direction {} not properly recognized'.format
        msg2 = 'step {} not properly recognized'.format
        solutions = [(0,'',0),
                     (0,'FORWARD',1),
                     (0.04224,'FORWARD',2),
                     (9.58567,'FORWARD',228)]
        for i,(obj,solution) in enumerate(zip(self.objects,solutions)):
            with self.subTest(Test_Object=i,reactioncoord=solution[0]):
                test = obj.reactioncoord
                sol = solution[0]
                self.assertEqual(test,sol ,msg0(sol))
            with self.subTest(Test_Object=i,reactioncoord=solution[1]):
                test = obj.direction
                sol = solution[1]
                self.assertEqual(test,sol ,msg1(sol))
            with self.subTest(Test_Object=i,reactioncoord=solution[2]):
                test = obj.step
                sol = solution[2]
                self.assertEqual(test,sol ,msg2(sol))

class TestLink202(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l202.txt')
        cls.refile = TEST_FILEDIR.joinpath('l202_regex.txt')
        cls.orifile = TEST_FILEDIR.joinpath('l202_orientations.txt')
        cls.dmatfile = TEST_FILEDIR.joinpath('l202_dmatrices.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link202(i) for i in txt.split(SMARK)]
        # Read and store the Solutions to the orientation tests
        with open(cls.orifile,'r') as F:
            txt = F.read()
        cls.orientations = txt.split(SMARK)

    def test_init(self):
        msg = 'Incorrect parsing of Link202'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,202)
            self.assertTrue(bool(obj.orientation),msg)

    def test_regex(self):
        msg = 're_orientation regex does not match properly'
        regex = Link202.re_orientation
        with open(self.refile) as F:
            txt = F.read()
        items = [i.split(MMARK) for i in txt.split(SMARK)]
        for i,(obj,solutions) in enumerate(zip(self.objects,items)):
            text = obj.text
            tests = regex.findall(text)
            for j,(test,sol) in enumerate(zip(tests,solutions)):
                with self.subTest(text=i,match=j):
                    self.assertEqual(test,sol,msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link202'
        obj = Link202('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,202)
        self.assertFalse(bool(obj.orientation),msg)

    def test_orientation(self):
        msg = 'Orientation not properly parsed'
        AtomCoords = Link202._AtomCoords
        with open(self.orifile) as F:
            txt = F.read()
        items = []
        for sol in txt.split(SMARK):
            lines = sol.split('\n')
            mat = []
            for line in lines:
                if line.strip():
                    values = line.strip().split()
                    a, b, c = int(values[0]),int(values[1]),int(values[2])
                    d, e, f = float(values[3]),float(values[4]),float(values[5])
                    mat.append(AtomCoords(a,b,c,d,e,f))
            items.append(mat)
        for i,(obj,solution) in enumerate(zip(self.objects,items)):
            test = obj.orientation
            with self.subTest(text=i):
                self.assertEqual(test,solution,msg)

    def test_DistanceMatrix(self):
        msg = 'DistanceMatrix not properly parsed'
        obj = self.objects[3]
        dmatrix_sol = []
        with open(self.dmatfile) as F:
            for line in F:
                if line.strip():
                    Aux = line.strip().split()
                    i0 = int(Aux.pop(0))
                    i1 = Aux.pop(0)
                    i2 = list(map(float, Aux))
                    dmatrix_sol.append([i0,i1,i2])
        self.assertTrue(dmatrix_sol == obj.DistanceMatrix,msg)

    def test_AtNum2Sym_error(self):
        msg = 'Obtained an unobtainable Mapping'
        Tests = self.objects
        with self.assertRaises(StopIteration,msg=msg):
            Tests[0].get_atom_mapping()
        with self.assertRaises(StopIteration,msg=msg):
            Tests[1].get_atom_mapping()
        with self.assertRaises(StopIteration,msg=msg):
            Tests[2].get_atom_mapping()

    def test_AtNum2Sym(self):
        msg = 'Incorrect Mapping Obtained'
        SolDict = {1:'H',8:'O',6:'C',7:'N'}
        TestDict = self.objects[3].get_atom_mapping()
        self.assertTrue(TestDict == SolDict,msg)

class TestLink502(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l502.txt')
        cls.re_energies_solutions = ['-2847.29438388','-2847.29924823',
                                     '-2847.29888359','-2847.29999372',
                                     '-2847.30039709','-2847.30030072',
                                     '-2847.30072337','-2847.30086336',
                                     '-2847.30091754','-2847.30086384',
                                     '-2847.30093740']
        cls.re_spin_solutions = [('0.7531', '0.7500'),
                                 ('0.7530', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500'),
                                 ('0.7529', '0.7500')]
        
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link502(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link502'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,502,msg)
            self.assertTrue(obj.energy is not None,msg)

    def test_regex_energy(self):
        msg = 're_EDone regex does not match properly'
        regex = Link502.re_EDone
        solutions = self.re_energies_solutions
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test == solution,msg)

    def test_regex_spin(self):
        msg = 're_spin regex does not match properly'
        regex = Link502.re_spin
        solutions = self.re_spin_solutions
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test == solution,msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link502'
        obj = Link502('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,502)

    def test_energy(self):
        msg = 'Energy value not properly read'
        solutions = [float(i) for i in self.re_energies_solutions]
        for obj,solution in zip(self.objects,solutions):
            test = obj.energy
            self.assertTrue(test == solution,msg)
    def test_spin(self):
        msg = 'S**2 values not properly read'
        solutions = [(float(i[0]),float(i[1])) for i in self.re_spin_solutions]
        for obj,solution in zip(self.objects,solutions):
            test = obj.spin
            self.assertTrue(test == solution,msg)

class TestLink601(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l601.txt')
        cls.regexfile = TEST_FILEDIR.joinpath('l601_regex.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link601(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link601'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,601,msg)
            self.assertTrue(obj.mulliken_heavy,msg)
            self.assertTrue(obj.mulliken,msg)

    def test_regex_MullikenAtoms(self):
        msg = 're_MullikenAtoms does not match the first group\n{}\n!=\n{}'
        msg = msg.format
        msg2 = 're_MullikenAtoms does not match the second group\n{}\n!=\n{}'
        msg2 = msg2.format
        regex = Link601.re_MullikenAtoms
        with open(self.regexfile,'r') as F:
            txt = F.read()
        samples = txt.split(SMARK)
        solutions = []
        for sample in samples:
            Aux = sample.split('\n## re_MullikenHeavy ##\n')[0]
            match0,match1 = Aux.split('\n## Match1 ##\n')
            match0 = match0.split('## Match0 ##\n')[-1]
            solutions.append((match0,match1))
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test[0] == solution[0],msg(test[0],solution[0]))
            self.assertTrue(test[1] == solution[1],msg2(test[1],solution[1]))
    def test_regex_MullikenHeavy(self):
        msg = 're_MullikenHeavy does not match the first group\n{}\n!=\n{}'
        msg = msg.format
        msg2 = 're_MullikenHeavy does not match the second group\n{}\n!=\n{}'
        msg2 = msg2.format
        regex = Link601.re_MullikenHeavy
        with open(self.regexfile,'r') as F:
            txt = F.read()
        samples = txt.split(SMARK)
        solutions = []
        for sample in samples:
            Aux = sample.split('\n## re_MullikenHeavy ##\n')[1]
            match0,match1 = Aux.split('\n## Match1 ##\n')
            match0 = match0.split('## Match0 ##\n')[-1]
            solutions.append((match0,match1))
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test[0] == solution[0],msg(test[0],solution[0]))
            self.assertTrue(test[1] == solution[1],msg2(test[1],solution[1]))

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link502'
        obj = Link601('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,601)

class TestLink716(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile        = TEST_FILEDIR.joinpath('l716.txt')
        cls.dipolefile      = TEST_FILEDIR.joinpath('l716_redipole.txt')
        cls.Thermofile      = TEST_FILEDIR.joinpath('l716_reThermo.txt')
        cls.EContribfile    = TEST_FILEDIR.joinpath('l716_reEContrib.txt')
        cls.IRSpectrumfile  = TEST_FILEDIR.joinpath('l716_reIRSpectrum.txt')
        cls.Frequenciesfile = TEST_FILEDIR.joinpath('l716_reFrequencies.txt')
        cls.freqtextfile    = TEST_FILEDIR.joinpath('l716_refreqtxt.txt')
        cls.freqdispfile    = TEST_FILEDIR.joinpath('l716_refreqdisplacements.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link716(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link716'
        attrs = Link716.__slots__
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,716,msg)
        for attr in attrs:
            self.assertTrue(getattr(self.objects[2],attr),msg)

    def test_regex_Thermo(self):
        msg = 're_Thermo does not match. \n{}\n!=\n{}'.format
        msg2 = 're_Thermo matches when it should not {}'.format
        regex = Link716.re_Thermo
        with open(self.Thermofile,'r') as F:
            txt = F.read()
        solutions = txt.split(SMARK)
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)
            if solution != '':
                test = test[0]
                self.assertTrue(test == solution,msg(test,solution))
            else:
                self.assertFalse(bool(test),msg2(test))
    def test_regex_EContrib(self):
        msg = 're_EContrib does not match. \n{}\n!=\n{}'.format
        msg2 = 're_EContrib matches when it should not {}'.format
        regex = Link716.re_EContrib
        with open(self.EContribfile,'r') as F:
            txt = F.read()
        solutions = txt.split(SMARK)
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)
            if solution != '':
                test = test[0]
                self.assertTrue(test == solution,msg(test,solution))
            else:
                self.assertFalse(bool(test),msg2(test))
    def test_regex_IRSpectrum(self):
        msg = 're_IRSpectrum does not match. \n{}\n!=\n{}'.format
        msg2 = 're_IRSpectrum matches when it should not {}'.format
        regex = Link716.re_IRSpectrum
        with open(self.IRSpectrumfile,'r') as F:
            txt = F.read()
        solutions = txt.split(SMARK)
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)
            if solution != '':
                test = test[0]
                self.assertTrue(test == solution,msg(test,solution))
            else:
                self.assertFalse(bool(test),msg2(test))
    def test_regex_Frequencies(self):
        msg = """re_Frequency does not properly match.'.format
        bool(
        {}
        )
        !=bool(
        {}
        )""".format
        msg2 = 'Frequency slice does not match. \n{}\n!=\n{}'.format
        regex = Link716.re_Frequencies
        with open(self.Frequenciesfile,'r') as F:
            txt = F.read()
        samples = txt.split(SMARK)
        solutions = []
        for sample in samples:
            if sample:
                Matches = sample.split('\n## Match ##\n')
                Aux = []
                for match in Matches:
                    Aux.append(tuple(match.split('\n')))
                solutions.append(Aux)
            else:
                solutions.append([])
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)
            self.assertTrue(bool(test) == bool(solution),msg(test,solution))
            for t,s in zip(test,solution):
                self.assertTrue(t == s, msg2(t,s))
    def test_regex_freq_text(self):
        msg = 'the text block does not match'
        regex = Link716.re_freq_text
        with open(self.freqtextfile,'r') as F:
            txt = F.read()
        solutions = txt.split(SMARK)
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test == solution,msg)
    def test_regex_freq_text(self):
        msg = 'the displacement: \n{} does not match:\n{}'.format
        regex = Link716.re_freq_displacements
        regex_txt = Link716.re_freq_text
        with open(self.freqdispfile,'r') as F:
            txt = F.read()
        items = txt.split(SMARK)
        solutions = []
        for item in items: 
            if item:
                solutions.append([s+'\n' for s in item.split('\n')])
            else:
                solutions.append([])
        for obj,solution in zip(self.objects,solutions):
            test = regex_txt.findall(obj.text)
            if test:
                subtests = regex.findall(test[0])
                for t,s in zip(subtests,solution): 
                    self.assertEqual(t,s,msg(test,solution))
            else:
                self.assertTrue(test == solution,msg(test,solution))

    def test_regex_dipole(self):
        msg = 're_dipole does not match. \n{}\n!=\n{}'
        msg = msg.format
        regex = Link716.re_dipole
        with open(self.dipolefile,'r') as F:
            txt = F.read()
        samples = txt.split(SMARK)
        solutions = []
        for sample in samples:
            solutions.append(tuple(sample.split('\n')))
        for obj,solution in zip(self.objects,solutions):
            test = regex.findall(obj.text)[0]
            self.assertTrue(test == solution,msg(test,solution))

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link716'
        obj = Link716('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,716)

class TestLink804(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l804.txt')
        cls.propfile = TEST_FILEDIR.joinpath('l804_properties.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link804(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link804'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,804)
            self.assertTrue(obj.MP2 is not None,msg)
            self.assertTrue(bool(obj.SpinComponents),msg)

    def test_regex_MP2(self):
        msg = 're_MP2 regex does not match properly'
        regex = Link804.re_MP2
        with open(self.propfile,'r') as F:
            txt = F.read()
        solutions = [lines.split('\n')[0] for lines in txt.split(SMARK)]
        items = [obj.text for obj in self.objects]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)[0]
            self.assertTrue(test == solution,msg)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link804'
        text = 'Link1:  Proceeding to internal job step number 13.'
        obj = Link804(text,asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,804)
        self.assertTrue(obj.MP2 is None,msg)
        self.assertFalse(bool(obj.SpinComponents),msg)

    def test_SpinComponents(self):
        msg = 'SpinComponents not properly parsed'
        SpinComponent = Link804._SpinComponent
        with open(self.propfile,'r') as F:
            txt = F.read().replace('D','E')
        items = [lines for lines in txt.split(SMARK)]
        for item,obj in zip(items,self.objects):
            sol = []
            lines = item.split('\n')[1:]
            for line in lines:
                a0,a1,a2 = line.split()
                sol.append(SpinComponent(a0,float(a1),float(a2)))
            self.assertTrue(obj.SpinComponents == sol,msg)

    def test_SCSCorr(self):
        msg = 'SCS correction not properly calculated'
        MockObject = Link804('',asEmpty=True)
        SpinComponent = Link804._SpinComponent
        MockObject.SpinComponents = [None,None,None]
        MockDict = dict(Name='',E=0,T=0)
        for item in [(0,0,0),(3,5,3),(1.5,10,1.5)]:
            for i,j in enumerate(item):
                MockDict['E'] = j
                MockObject.SpinComponents[i] = SpinComponent(**MockDict)
            test = MockObject.get_SCScorr()
            aa,ab,bb = item
            sol = (aa+bb)/3.0 +(6.0*ab)/5.0
            self.assertTrue(test == sol,msg)

    def test_MP2(self):
        msg = 'MP2 Energy not properly parsed'
        with open(self.propfile,'r') as F:
            txt = F.read().replace('D','E')
        solutions = [float(lines.split('\n')[0]) for lines in txt.split(SMARK)]
        for sol,obj in zip(solutions,self.objects):
            self.assertTrue(obj.MP2 == sol,msg)

class TestLink913(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.testfile = TEST_FILEDIR.joinpath('l913.txt')
        cls.MP4file = TEST_FILEDIR.joinpath('l913_MP4.txt')
        cls.ccsdtfile = TEST_FILEDIR.joinpath('l913_ccsdt.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link913(i) for i in txt.split(SMARK)]

    def test_init(self):
        msg = 'Incorrect parsing of Link913'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,913)

    def test_regex_MP4(self):
        msg = 're_MP4 regex does not match properly'
        regex = Link913.re_MP4
        with open(self.MP4file) as F:
            txt = F.read()
        solutions = [i.split('\n') for i in txt.split(SMARK)]
        items = [obj.text for obj in self.objects]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            # Test only for the last match as it is the only one stored
            self.assertTrue(test[-1] == solution[-1],msg)

    def test_regex_ccsdt(self):
        msg_match = 're_ccsdt regex does not match properly\n{} != {}'
        msg_nomatch = 're_ccsdt regex does not find a match properly'
        regex = Link913.re_ccsdt
        with open(self.ccsdtfile) as F:
            txt = F.read()
        solutions = [i for i in txt.split(SMARK)]
        items = [obj.text for obj in self.objects]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            if test: # Test matches
                self.assertTrue(test[0] == solution,msg_match.format(test[0],solution))
            else: # Test non-matches, although the test is within the condition
                self.assertFalse(bool(test),msg_nomatch)

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link913'
        obj = Link913('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,913)

    def test_MP4(self):
        msg = 'MP4 energy not properly read'
        with open(self.MP4file) as F:
            txt = F.read().replace('D','E')
        solutions = [float(lines.split('\n')[-1].split()[-1])
                                        for lines in txt.split(SMARK)]
        for sol,obj in zip(solutions,self.objects):
            self.assertTrue(obj.MP4 == sol,msg)

    def test_CCSDT(self):
        msg = 'CCSDT energy not properly read'
        with open(self.ccsdtfile,'r') as F:
            txt = F.read().replace('D','E')
        solutions = [float(lines.strip()) if lines.strip() else None
                                        for lines in txt.split(SMARK)]
        for sol,obj in zip(solutions,self.objects):
            self.assertTrue(obj.CCSDT == sol,msg)

class TestLink914(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ES_mark = '\n## ExcitedState ##\n'
        cls.testfile = TEST_FILEDIR.joinpath('l914.txt')
        cls.propertiesfile = TEST_FILEDIR.joinpath('l914_properties.txt')
        # Read and store the Examples
        with open(cls.testfile,'r') as F:
            txt = F.read()
        cls.objects = [Link914(i) for i in txt.split(SMARK) if i.strip()]
        with open(cls.propertiesfile,'r') as F:
            txt = F.read()
        cls.solutions = [slice.split(ES_mark) for slice in txt.split(SMARK) if slice.strip()]

    def test_init(self):
        msg = 'Incorrect parsing of Link914'
        for obj in self.objects:
            self.assertTrue(bool(obj.text),msg)
            self.assertEqual(obj.number,914)

    def test_regex_ExcitedState(self):
        msg = 're_ExcitedState regex does not match {} with {}'.format
        regex = Link914.re_ExcitedState
        solutions = []
        for sol in self.solutions:
            aux = []
            for ES in sol:
                aux.append(tuple(ES.split('\n')[0].strip().split()))
            solutions.append(aux)
        items = [obj.text for obj in self.objects]
        for txt,solution in zip(items,solutions):
            test = regex.findall(txt)
            # Test only for the last match as it is the only one stored
            self.assertTrue(tuple(test) == tuple(solution),msg(test,solution))

    def test_regex_transition(self):
        msg = 're_transition regex does not match {} with {}'.format
        regex = Link914.re_transition
        # >> obj = solutions[0] = [[('111A','->','222B','-0.00000'),(...)], ...]
        # >> ES  = solutions[0][0] = [('111A','->','222B','-0.00000'),(...)]
        # >> orb = solutions[0][0][0] = ('111A','->','222B','-0.00000')
        solutions = []
        for sol in self.solutions:
            aux = []
            for ES in sol:
                transitions = []
                for transition in ES.split('\n')[1:]:
                    if transition.strip():
                        transitions.append(tuple(transition.strip().split()))
                aux.append(transitions)
            solutions.append(aux)
        items = [obj.text for obj in self.objects]
        for txt,solution in zip(items,solutions):
            tests = regex.findall(txt)
            for test,sol in zip(tests,list(chain.from_iterable(solution))):
                self.assertTrue(test == sol,msg(test,sol))

    def test_init_empty(self):
        msg = 'Incorrect empty initialization of Link913'
        obj = Link914('Some Text',asEmpty=True)
        self.assertFalse(bool(obj.text),msg)
        self.assertEqual(obj.number,914)

    def test_locate_ExcitedStates(self):
        msg = 'ExcitedStates not read properly {} does not match {}'.format
        solutions = []
        for sol in self.solutions:
            aux = []
            for ES in sol:
                a,_,b,c,d,e = ES.split('\n')[0].strip().split()
                a = int(a)
                b,c,d,e = tuple(map(float,(b,c,d,e)))
                aux.append((a,b,c,d,e))
            solutions.append(aux)
        for solution,obj in zip(solutions,self.objects):
            for sol,ES in zip(solution,obj.excitedstates):
                test = tuple(ES[:-1])
                self.assertTrue(sol == test ,msg(sol,test))

    def test_extract_transitions(self):
        msg = 'Wrong Transition Extraction: {1} does not match {0}'.format
        L914 = Link914('',asEmpty=True)
        txt = """
             107B ->166B        0.02496
             165B ->167B        0.02444
             126B <-166B       -0.01650
             Sometexthere
             ## Split Here ##
             165 -> 167        0.02444
             126 <- 166       -0.01650

             """
        solutions = [[('107B','166B',float('0.02496' ),False),
                      ('165B','167B',float('0.02444' ),False),
                      ('166B','126B',float('-0.01650'),True)],
                     [( '165', '167',float('0.02444' ),False),
                      ( '166', '126',float('-0.01650'),True)]]
        for sol,intxt in zip(solutions,txt.split(SMARK)):
            transitions = L914._extract_transitions(intxt)
            for sol,test in zip(sol,transitions):
                self.assertEqual(tuple(test),sol,msg(test,sol))


if __name__ == '__main__':
    unittest.main()



"""
One of the two core modules of pyssian. This module contains the classes and
parsers for the outputs of the different types of links. Only the classes
registered with the appropiate decorator are considered for parsing.
"""

import warnings
from collections import namedtuple
from itertools import cycle
import re

# Auxiliar Documenting Decorators
def Populates(*attributes, defaults=None):
    """
    Helper decorator function to document _locate* methods

    Parameters
    ----------
    *attributes : str
        names of the attributes populated by the method.
    default : list
        list of default values in the same order as the attributes. If none
        provided no "default message" is added.

    Returns
    -------
    subdecorator
        returns a decorator that attaches a message constructed
        with the attributes and defaults specified
    """

    if len(attributes) > 0:
        if defaults is None:
            Out = [f'- {attr}' for attr in attributes]
        elif len(defaults) > 1:
            Out = [f'- {attr} ({default})'
                    for attr, default in zip(attributes,defaults)]
        else:
            Out = [f'- {attr} ({defaults[0]})' for attr in attributes]
        Out.insert(0,'Populates the attributes:')
        Out.append('')
    else:
        Out = [f'Populates the attribute: **{attributes[0]}**',]
    def subdecorator(f):
        match = re.search(r'^[\s]*',f.__doc__)
        lpadding = '\n'
        if match:
            lpadding += match[0]
        f.__doc__ += lpadding + lpadding.join(Out) + '\n'
        return f
    return subdecorator
def SilentFail(f):
    f.__doc__ += '(Fails Silently).'
    return f

#Base Class
class LinkJob(object):
    """LinkJob Base class. Represents the output of a lxxxx.exe.

    Parameters
    ----------
    text : list
        text that corresponds to the output of the lxxxx.exe
    number : int
        Integrer that represents the type of the LinkJob (the default is None).

    Attributes
    ----------
    Register : dict
        A dict where key corresponds to an int, and value corresponds to the
        registered subclass class through @RegisterLinkJob
    number

    text

    """
    __slots__ = ('text','number')
    _token = 'Base'
    Register = {}
    def __init__(self,text,number=None,asEmpty=False):
        self.number = number
        if asEmpty:
            text = ''
        self.text = text
    def __repr__(self):
        cls = type(self).__name__
        return f'<Link {self.number} of class {cls}>'
    def __str__(self):
        return ''.join(self.text)

    def _text_iterbyline(self):
        return [line+'\n' for line in self.text.split('\n')].__iter__()

    @classmethod
    def as_empty(cls,text,number=None):
        instance = cls(text,asEmpty=True)
        if instance.number is None:
            instance.number = cls._token
        return instance

class GeneralLinkJob(LinkJob):
    """
    Subclass of Linkjob to store the information of any Link that does not
    have a specific parser.

    Parameters
    ----------

    text : str
        text that corresponds to the output of the lxxxx.exe
    number : int
        Integrer that represents the type of the LinkJob. (defaults to None)
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------

    number

    text

    """

    _token = 'General'

    __slots__ = ('number', 'text')

    re_enter = re.compile(r'(?:Enter.*l)([0-9]{1,4})(?:\.exe)')

    def __init__(self,text,number=None,asEmpty=False):
        cls = self.__class__
        if number is None:
            Match = cls.re_enter.findall(text)
            if Match:
                number = int(Match[0])
            else:
                number = -1
        if asEmpty:
            text = ''
        super().__init__(text,number)

def RegisterLinkJob(cls):
    """Decorator for the LinkJob subclasses, adds the class to the
    LinkJob.Register dictionary with the _token as key.

    Parameters
    ----------
    cls : LinkJob
        LinkJob subclass identified with the cls._token attribute. Which is
        normally an integrer between 1 and 9999.

    """
    key = cls._token
    if key in LinkJob.Register:
        msg = '\nWarning! overwriting :\n{} with {}\n'.format
        warnings.warn(msg(LinkJob.Register[key],cls))
    LinkJob.Register[key] = cls
    return cls

@RegisterLinkJob
class Link1(LinkJob):
    """
    Representation and parser for the output of l1.exe,it also holds
    information about the InternalJob setup.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l1.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    info : namedtuple
        tuple that contains information needed to instantiate an InternalJob.
    commandline : str
        contains the information of the gaussian commands
        (defaults to an empty string).
    nprocs : int
        contains the number of the %nprocshared link0 option
    mem : str
        contains the right hand side of the %mem link0 option
    link0 : list
        list of lines containing the link0 specifications.
    IOps : list
        list of lines containing the IOps that appear below the command line
    text
    number
    """

    _token = 1

    re_link0 = re.compile(r'\%.*')
    re_commandline = re.compile(r'(?:\-{10,}\n)(.*?#[\s\S]*?)(?:\-{10,}\n)')
    re_IOps = re.compile(r'.*\;')
    re_internaljob = re.compile(r'Link1\:.*internal.*job.*([0-9]{1,3})\.')

    InternalJobInfo = namedtuple('InternalJobInfo','number type new_InternalJob')

    def __init__(self,text,asEmpty=False):
        super().__init__(text,1)
        self.info = None
        self.commandline = ''
        self.nprocs = None
        self.mem = None
        self.link0 = []
        self.IOps = []
        if asEmpty:
            self._locate_internaljob()
            self.text = '' # clear text
        else:
            self._locate_link0()
            self._locate_commandline()
            self._locate_IOps()
            self._locate_internaljob()

    @Populates('info',defaults=['If none is found defaults to 1'])
    def _locate_internaljob(self):
        """
        Uses regex expressions compiled as class attributes to search the number
        of the "current" internal job at the file.
        """
        cls = self.__class__
        Match = cls.re_internaljob.findall(self.text)
        if Match:
            JobNumber = int(Match[0])
            New_InternalJob = True
        else:
            JobNumber = 1
            New_InternalJob = False
        jobtype = self._guess_type()
        self.info = cls.InternalJobInfo(JobNumber, jobtype, New_InternalJob)

    @Populates('nprocs','mem', 'link0', defaults=['Defaults to None',
    'Defaults to None','If none is found defaults to an empty list'])
    def _locate_link0(self):
        """
        Uses regex expressions compiled as class attributes to search the link0
        options (specified with a % at the start of the option).
        """
        link0_opt = []
        cls = self.__class__
        Match = cls.re_link0.findall(self.text)
        if Match:
            for item in Match:
                if 'nprocshared' in item:
                    self.nprocs = int(item.strip().split('=')[-1])
                elif 'mem' in item:
                    self.mem = item.strip().split('=')[-1]
                else:
                    link0_opt.append(item[1:])
        self.link0 = link0_opt

    @Populates('commandline')
    @SilentFail
    def _locate_commandline(self):
        """
        Uses regex expressions compiled as class attributes to search the
        gaussian commands, which start with a "#".
        """
        cls = self.__class__
        Match = cls.re_commandline.findall(self.text)
        if Match:
            lines = Match[0].split('\n')
            commandline = ''.join([line.lstrip() for line in lines])
            self.commandline = commandline

    @Populates('IOps')
    @SilentFail
    def _locate_IOps(self):
        """
        Uses regex expressions compiled as class attributes to search IOps (see
        Gaussian Manual).
        """
        cls = self.__class__
        Match = cls.re_IOps.findall(self.text)
        if Match:
            self.IOps = [line for line in Match]

    def _guess_type(self):
        """
        Guesses the type of Job depending on the keywords in the
        self.commandline.

        Returns
        -------
        str
            {Constrained Optimization, Optimization,Frequency Calculation, 
            Unidentified, Linked} the Linked type is only assigned externally
        """
        target = self.commandline.lower()
        if 'modredundant' in target:
            jobtype = 'Constrained Optimization'
        elif 'opt' in target:
            jobtype = 'Optimization'
        elif 'freq' in target:
            jobtype = 'Frequency Calculation'
        else:
            jobtype = 'Unidentified'
        return jobtype

@RegisterLinkJob
class Link101(LinkJob):
    """
    Representation and parser for the output of l101.exe

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l101.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults is False)

    Attributes
    ----------
    charge : int
    spin : int
    """
    __slots__ = ('charge','spin')

    _token = 101

    re_charge = re.compile(r'(?:Charge\s=\s*)(-{0,1}[0-9]*)')
    re_spin = re.compile(r'(?:Multiplicity\s=\s*)([0-9]*)')

    def __init__(self,text,asEmpty=False):
        self.charge = None
        self.spin = None
        if asEmpty:
            super().__init__('',101)
        else:
            super().__init__(text,101)
            self._locate_charge_spin()
    @Populates('charge','spin')
    @SilentFail
    def _locate_charge_spin(self):
        """
        Searches the charge and spin.
        """
        cls = self.__class__
        charge = cls.re_charge.findall(self.text)
        spin = cls.re_spin.findall(self.text)
        if charge:
            self.charge = int(charge[0])
        if spin:
            self.spin = int(spin[0])

@RegisterLinkJob
class Link103(LinkJob):
    """
    Representation and parser for the output of l103.exe. Corresponds to a
    single step of a Berny optimization.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l103.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    mode : str
        'Init', 'Iteration' or 'End'
    state : str
        Either 'Optimized', 'Non-optimized' or 'Initial'
    convergence : list
        List of namedtuples with fields Item,Value,Threshold,Converged.
    parameters : list
        List of namedtuples with fields Name,Definition,Value,Derivative.
    derivatives : list
        List of namedtuples with fields Var,old,dEdX,dXl,dXq,dXt,new.
    stepnumber
    scanpoint
    text
    number

    """

    __slots__ = ('mode','state','convergence','parameters','derivatives',
                'stepnumber','scanpoint')

    _token = 103

    re_init = re.compile(r'Initialization\spass\.')
    re_end = re.compile(r'(Optimization\sstopped\.)|(Optimization completed\.)')
    re_parameters = re.compile(r'\!.*\!')
    re_derivatives = r'([R|D|X|A][0-9]*)'+r'\s*([-]?[0-9]*\.[0-9]{0,5})'*6
    re_derivatives = re.compile(re_derivatives)
    re_convergence = r'((?:Maximum|RMS)\s*(?:Force|Displacement))'
    re_convergence += r'\s*([0-9]{0,3}\.[0-9]*)\s*([0-9]{0,3}\.[0-9]*)\s*(YES|NO)'
    re_convergence = re.compile(re_convergence)
    re_stepnum = re.compile(r'Step\s*number\s*([0-9]*)\s*out\s*of')
    re_scanpoint = re.compile(r'scan\spoint\s*([0-9]*)\s*out\s*of')

    _Parameter = namedtuple("Parameter", "Name Definition Value Derivative")
    _ConverItem = namedtuple("ConverItem", "Item Value Threshold Converged")
    _Derivative = namedtuple("Derivative", "Var old dEdX dXl dXq dXt new")

    def __init__(self,text,asEmpty=False):
        self.mode = None
        self.state = None
        self.convergence = []
        self.parameters = []
        self.derivatives = []
        self.stepnumber = None
        self.scanpoint = None
        if asEmpty:
            super().__init__('',103)
        else:
            super().__init__(text,103)
            self._locate_mode()
            self._locate_parameters()
            self._locate_derivatives()
            self._locate_convergence()
            self._locate_numbers()

    @Populates('mode',defaults=['Iteration'])
    def _locate_mode(self):
        """
        Uses regex expressions compiled as class attributes to find
        If the Link corresponds to an (Init)ialization pass, (Iteration)
        or (End) converged or not.
        """
        cls = self.__class__ # Alias
        IsInit = cls.re_init.findall(self.text)
        IsEnd  = cls.re_end.findall(self.text)
        if IsInit:
            self.mode = 'Init'
        elif IsEnd:
            self.mode = 'End'
        else:
            self.mode = 'Iteration'

    @Populates('state','parameters')
    @SilentFail
    def _locate_parameters(self):
        """
        Uses regex expressions compiled as class attributes to find
        the parameter table if exists.
        """
        cls = self.__class__
        Match = cls.re_parameters.findall(self.text)
        Parameters = []
        if Match:
            state = Match.pop(0)
            # '! {Initial/Optimized/Non-Optimized} Parameters !'
            self.state = state.replace('!','').strip().split()[0]
            _ = Match.pop(0) # Discard Units
            _ = Match.pop(0) # Discard Table Headings
            for line in Match:
                Columns = line.strip().split()
                Name = Columns[1]
                Definition = Columns[2]
                Value = Columns[3]
                if self.mode == "Init":
                    Derivative = Columns[-2]
                else:
                    Derivative = float(Columns[-2])
                Par = cls._Parameter(Name, Definition,
                                    Value, Derivative)
                Parameters.append(Par)
            self.parameters = Parameters

    @Populates('state','derivatives')
    @SilentFail
    def _locate_derivatives(self):
        """
        Uses regex expressions compiled as class attributes to find
        the derivatives table if exists.
        """
        cls = self.__class__
        Match = cls.re_derivatives.findall(self.text)
        items = []
        if Match:
            for m in Match:
                Var, old, dEdX, dXl, dXq, dX, new = m
                item = cls._Derivative(Var, old, dEdX, dXl, dXq, dX, new)
                items.append(item)
            self.derivatives = items

    @Populates('convergence')
    @SilentFail
    def _locate_convergence(self):
        """
        Uses regex expressions compiled as class attributes to find
        the convergence parameters.
        """
        cls = self.__class__
        Match = cls.re_convergence.findall(self.text)
        items = []
        if Match:
            for m in Match:
                Name = ' '.join(m[0].split())
                val = float(m[1])
                threshold = float(m[2])
                isConverged = m[3] == "YES"
                item = cls._ConverItem(Name,val,threshold,isConverged)
                items.append(item)
            self.convergence = items

    @Populates('stepnumber','scanpoint')
    @SilentFail
    def _locate_numbers(self):
        """
        Uses regex expressions compiled as class attributes to find
        the current step number (of an optimization) and scan number
        (of an scan calculation).
        """
        cls = self.__class__
        Match = cls.re_stepnum.findall(self.text)
        if Match:
            self.stepnumber = int(Match[0])
        elif self.mode == 'Init':
            self.stepnumber = 0
        Match = cls.re_scanpoint.findall(self.text)
        if Match:
            self.scanpoint = int(Match[0])

    def print_convergence(self):
        """ Prints the convergence Table formatted """
        Format = '{0: <22s}\t{1:.6f}\t{2:.6f}\t{3}'.format
        try:
            Out = [Format(*i) for i in self.convergence]
        except ValueError as e:
            print("Non numeric value found in the convergence values")
            Format_str='{0: <22s}\t{1}\t{2}\t{3}'.format
            Out = [Format_str(*i) for i in self.convergence]
        print('\n'.join(Out))

@RegisterLinkJob
class Link120(LinkJob):
    """
    Representation and parser for the output of l120.exe. Corresponds to a
    the energy calculations when using ONIOM.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l103.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    energy
    energy_partitions

    """
    __slots__ = ('energy', 'energy_partitions')

    _token = 120

    re_energy = r'extrapolated energy\s=\s*(-?[0-9]*\.[0-9]*)'
    re_energy = re.compile(re_energy)
    re_energy_partitions = r'gridpoint\s*([0-9]*)\smethod:\s*([a-zA-Z]*)\s*system:\s*([a-zA-Z]*)\s*energy:\s*(-?[0-9]*\.[0-9]*)'
    re_energy_partitions = re.compile(re_energy_partitions)
    _EnergyPartition = namedtuple('EnergyPartition','gridpoint level model energy')

    def __init__(self,text,asEmpty=False):
        self.energy = None
        self.energy_partitions = []
        if asEmpty:
            super().__init__('',120)
        else:
            super().__init__(text,120)
            self._locate_energies()

    @Populates('energy','energy_partitions',defaults=['None','[]'])
    @SilentFail
    def _locate_energies(self):
        """
        Uses regex expressions compiled as class attributes to find
        the energy and the different energy_partitions.
        """
        cls = self.__class__
        match = cls.re_energy.findall(self.text)
        if match:
            self.energy = float(match[0])

        EnergyPartition = cls._EnergyPartition
        match = cls.re_energy_partitions.findall(self.text)
        if not match:
            return 
        self.energy_partitions = [EnergyPartition(int(p),level,model,float(energy))
                                  for p,level,model,energy in match]

@RegisterLinkJob
class Link123(LinkJob):
    """
    Representation and parser for the output of l123.exe. Corresponds to irc
    predictor-corrector steps and may hold the orientation (Geometry) at a given
    step.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l202.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    orientation : list
        List of namedtuples containing the geometry of the calculated compound.
    step : int
        step number of along the irc calculation.
    direction : str
        FORWARD or REVERSE
    reactioncoord : float
       Value of the reaction coordinate at the current step.

    text
    number

    """
    __slots__ = ('orientation', 'direction', 'step', 'reactioncoord')

    _token = 123

    re_orientation =  r'^\s*([0-9]{1,4})\s*([0-9]{1,3})\s*'
    re_orientation += r'(\-?[0-9]+\.[0-9]*)'
    re_orientation += r'\s*(\-?[0-9]+\.[0-9]*)'*2
    re_orientation = re.compile(re_orientation, re.MULTILINE)

    re_reactioncoord =  r'NET REACTION COORDINATE UP TO THIS POINT\s*='
    re_reactioncoord += r'\s*([0-9]*\.[0-9]*)'
    re_reactioncoord = re.compile(re_reactioncoord)

    re_step = r'Point Number\s*([0-9]*)\s*in\s([A-Z]*)\spath direction'
    re_step = re.compile(re_step)

    re_iscomplete = r'(Calculation of (REVERSE|FORWARD) path complete.)'
    re_iscomplete = re.compile(re_iscomplete)

    re_totalsteps = r'\#\sOF\sPOINTS\sALONG\sTHE\sPATH\s*=\s*([0-9]*)'
    re_totalsteps = re.compile(re_totalsteps)

    _AtomCoords = namedtuple('AtomCoords',
                            'CenterNum AtomicNum AtomType X Y Z')

    def __init__(self,text,asEmpty=False):
        self.orientation = []
        self.direction = ''
        self.step = 0
        self.reactioncoord = 0
        if asEmpty:
            super().__init__('',123)
        else:
            super().__init__(text,123)
            self._locate_orientation()
            self._locate_irc_step()

    @Populates('orientation')
    @SilentFail
    def _locate_orientation(self):
        """
        Uses regex expressions compiled as class attributes to find
        the geometry of the molecule.
        """
        cls = self.__class__
        AtomCoords = cls._AtomCoords
        match = cls.re_orientation.findall(self.text)
        if match:
            for line in match:
                Atom = [int(line[0]),int(line[1]),None]
                Atom.extend([float(i) for i in line[2:]])
                self.orientation.append(AtomCoords(*Atom))

    @Populates('reactioncoord','step','direction')
    @SilentFail
    def _locate_irc_step(self):
        """
        Uses regex expressions compiled as class attributes to find
        the current step info of the irc path.
        """
        cls = self.__class__
        match = cls.re_reactioncoord.findall(self.text)
        if match:
            self.reactioncoord = float(match[0])
        match = cls.re_step.findall(self.text)
        if match:
            self.step = int(match[0][0])
            self.direction = match[0][1]
        else:
            iscomplete = cls.re_iscomplete.findall(self.text)
            if iscomplete:
                self.direction = iscomplete[0][1]
                totalsteps = int(cls.re_totalsteps.findall(self.text)[0])
                self.step = totalsteps + 1

    def print_orientation(self):
        """
        Displays in console the geometry with a format similar to gaussian.
        """
        Format = '{0}\t{1}\t{3: 0.6f}\t{4: 0.6f}\t{5: 0.6f}\n'
        Header = 'Center\tAtNum\t{0:^9s}\t{1:^9s}\t{2:^9s}'.format('X','Y','Z')
        print(Header)
        Out = [Format.format(*i) for i in self.orientation]
        print(''.join(Out))

@RegisterLinkJob
class Link202(LinkJob):
    """
    Representation and parser for the output of l202.exe. Holds the
    orientation (Geometry) at a given step.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l202.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    orientation : list
        List of namedtuples containing the geometry of the calculated compound.
    DistanceMatrix : list
        For each item the postitions 0 and 1 correspond to the Atom ID in the
        structure and the Atomic Number/Symbol respectively. Position 1 holds
        a list with the distances with respect to all previous items.
    text
    number

    """
    __slots__ = ('orientation', 'DistanceMatrix')

    _token = 202

    re_orientation = r'(?:.*(?:Standard|Input)\sorientation\:.*\n'
    re_orientation += r'\s?\-+\n.*\n.*\n\s?\-+\n)'
    re_orientation += r'[\s\S]*?'
    re_orientation += r'(?:\s?\-+\n)'
    re_orientation = re.compile(re_orientation)
    re_number = re.compile('[0-9]')

    _AtomCoords = namedtuple('AtomCoords',
                            'CenterNum AtomicNum AtomType X Y Z')

    def __init__(self,text,asEmpty=False):
        self.orientation = []
        self.DistanceMatrix = []
        if asEmpty:
            super().__init__('',202)
        else:
            super().__init__(text,202)
            self._locate_orientation()
            self._locate_DistanceMatrix()

    @Populates('orientation')
    @SilentFail
    def _locate_orientation(self):
        """
        Uses regex expressions compiled as class attributes to find
        the geometry of the molecule.
        """
        cls = self.__class__
        AtomCoords = cls._AtomCoords
        match = cls.re_orientation.findall(self.text)
        if match:
            # If Input and Standard orientation coexist
            # take the last one (Standard, generally)
            lines = [i for i in match[-1].split('\n') if i.strip()]
            _ = lines.pop(0) # Discard "Input orientation" line
            # Discard "Header"
            for i in range(4):
                _ = lines.pop(0)
            # Discard Last Line that
            _ = lines.pop(-1)
            for line in lines:
                Aux = line.split()
                Atom = [int(i) for i in Aux[0:3]]
                Atom.extend([float(i) for i in Aux[3:]])
                self.orientation.append(AtomCoords(*Atom))

    @Populates('DistanceMatrix')
    @SilentFail
    def _locate_DistanceMatrix(self):
        """
        Looks for the keyword 'Distance matrix' and reads the
        Input orientation table that follows
        """
        Iterator = self._text_iterbyline()
        cls = self.__class__
        for line in Iterator:
            if "Distance matrix" in line:
                break
        else: # Fail Silently
            return
        for _ in range(2):
            line = next(Iterator)
        Items = []
        while cls.re_number.match(line.strip()[0]):
            # while the first non blank character is a number
            Atom = line.strip().split()
            Item = []
            if '.' not in line or cls.re_number.match(Atom[1]):
                line = next(Iterator)
                continue
            Item.append(int(Atom[0]))
            Item.append(Atom[1])
            Item.append(list(map(float,Atom[2:])))
            Items.append(Item)
            line = next(Iterator)
        DicMatrix = {i[0]:['X',[]] for i in Items}
        for i in Items:
            Atom = DicMatrix[i[0]]
            if Atom[0] == "X":
                Atom[0] = i[1]
            else:
                if Atom[0] != i[1]:
                    raise RuntimeError('''Two atoms with the same Center number
                                     and different Symbol have been found ''')
            Atom[1].extend(i[2])
        self.DistanceMatrix = []
        for Id, (Sym, Dist) in DicMatrix.items():
            self.DistanceMatrix.append([Id,Sym,Dist])

    def print_orientation(self):
        """
        Displays in console the geometry with a format similar to gaussian.
        """
        Format = '{0}\t{1}\t{3: 0.6f}\t{4: 0.6f}\t{5: 0.6f}\n'.format
        Header = 'Center\tAtNum\t{0:^9s}\t{1:^9s}\t{2:^9s}'.format('X','Y','Z')
        print(Header)
        Out = [Format(*i) for i in self.orientation]
        print(''.join(Out))
    def get_atom_mapping(self):
        """
        Returns a dictionary that relates the atomic number with the symbol
        based on the information contained in the orientation and distance
        matrix.

        Returns
        -------
        Out
            Dictionary with Atomic Number to Symbol mapping
        """
        Out = dict()
        if not self.orientation: # In case it was parched
            raise RuntimeError('Attempting to access a non-existing orientation')
        atoms = [(atom[0],atom[1]) for atom in self.orientation ]
        _iter = self._text_iterbyline()
        for line in _iter:
            if 'Distance matrix' in line:
                break
        try:
            _ = next(_iter)
        except StopIteration:
            raise StopIteration('Attempting to get a mapping from a l202 without Distance matrix')
        for AtId,AtNum in sorted(atoms,key=lambda x:float(x[0])):
            line = next(_iter)
            Aux = line.strip().split()[:2]
            _, AtSym = Aux
            Out[AtNum] = AtSym
        return Out

#@RegisterLinkJob
class Link301(LinkJob):
    """
    Parser for the output of l301.exe, each execution displays
    information about solvation model. Holds information about
    Cavity Surface Area, and Cavity Volume. In development.
    """
    _token = 301

@RegisterLinkJob
class Link502(LinkJob):
    """
    Representation and parser for the output of l502.exe. Corresponds to a
    iterative calculation of the SCF cycles.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l502.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    energy : float
        Final Potential Energy of the SCF cycles.
    spin : tuple[float,float]
        S**2 (before, after) annihiliation
    """
    __slots__ = ('energy','spin')

    _token = 502

    re_EDone = re.compile(r'SCF\sDone\:\s*E\(.*\).*=\s*(\-?[0-9]*\.[0-9]+)')
    re_spin = re.compile('S\*\*2 before annihilation\s*(-?[0-9]*\.[0-9]*),\s*after\s*(-?[0-9]*\.[0-9]*)')

    def __init__(self,text,asEmpty=False):
        self.energy = None
        if asEmpty:
            super().__init__('',502)
        else:
            super().__init__(text,502)
            self._locate_energy()
            self._locate_spin()

    @Populates('energy')
    @SilentFail
    def _locate_energy(self):
        """
        Uses regex expressions compiled as class attributes to find the Energy
        rported as 'SCF Done:' and reads the potential energy.
        """
        cls = self.__class__
        Match = cls.re_EDone.findall(self.text)
        if Match:
            self.energy = float(Match[0])

    @Populates('spin')
    @SilentFail
    def _locate_spin(self):
        """
        Uses regex expressions compiled as class attributes to find the Energy
        rported as 'SCF Done:' and reads the potential energy.
        """
        cls = self.__class__
        Match = cls.re_spin.findall(self.text)
        if Match:
            before,after = Match[0]
            self.spin = float(before),float(after)
    

@RegisterLinkJob
class Link508(Link502):
    """
    Representation and parser for the output of l508.exe. Corresponds to a
    iterative calculation of the SCF using a quadratic convergence algorithm.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l508.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    energy : float
        Final Potential Energy of the SCF cycles.
    spin : tuple[float,float]
        S**2 (before, after) annihiliation
    """
    __slots__ = ('energy','spin')

    _token = 508

    def __init__(self,text,asEmpty=False):
        super().__init__(text,asEmpty)
        self.number = 508

@RegisterLinkJob
class Link601(LinkJob):
    """
    Representation and parser for the output of l601.exe, 
    each does a population analysis using SCF densities. Holds information 
    about Mulliken Charges, {Dipole, Quadrupole, Traceless Quadrupole, Octapole 
    and Hexadecapole} moments, spin densities... etc. 
    (Currently only Mulliken Charges implemented)

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l508.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    mulliken_heavy : list
        list of mulliken charges condensed to H atoms. None if not properly parsed.
    mulliken : list
        list of mulliken charges. None if not properly parsed.
    """
    re_MullikenAtoms =  r'(?:Mulliken charges( and spin densities)?.*\n.*\n)'# Start
    re_MullikenAtoms += r'([\s\S]*)' # Body
    re_MullikenAtoms += r'(?:\n.Sum of Mulliken charges.*\n)' # End
    re_MullikenAtoms = re.compile(re_MullikenAtoms)
    re_MullikenHeavy =  r'(?:Mulliken charges( and spin densities)? with hydrogens.*\n.*\n)'
    re_MullikenHeavy += r'([\s\S]*?)'
    re_MullikenHeavy += r'(?:\n.[a-zA-Z].*\n)'
    re_MullikenHeavy = re.compile(re_MullikenHeavy)
    
    _token = 601

    def __init__(self,text,asEmpty=False):
        self._Atom = namedtuple('Atom','number symbol charge spin')
        self.mulliken_heavy = None
        self.mulliken = None
        if asEmpty:
            super().__init__('',601)
        else:
            super().__init__(text,601)
            self._locate_MullikenHeavy()
            self._locate_MullikenAtoms()

    @Populates('mulliken_heavy')
    @SilentFail
    def _locate_MullikenHeavy(self):
        """
        Uses regex expressions compiled as class attributes to find the
        Mulliken charges and spin condensed to heavy atoms.
        """
        cls = self.__class__
        Match = cls.re_MullikenHeavy.findall(self.text)
        if Match:
            lines =  Match[0][1].split('\n')
            if Match[0][0]:
                Atom = lambda x: self._Atom(int(x[0]),x[1],float(x[2]),float(x[3]))
            else:
                Atom = lambda x: self._Atom(int(x[0]),x[1],float(x[2]),spin=None)
            self.mulliken_heavy = [Atom(line.strip().split())
                                    for line in lines if line.strip()]

    @Populates('mulliken')
    @SilentFail
    def _locate_MullikenAtoms(self):
        """
        Uses regex expressions compiled as class attributes to find the
        atomic Mulliken charges.
        """
        cls = self.__class__
        Match = cls.re_MullikenAtoms.findall(self.text)
        if Match:
            lines =  Match[0][1].split('\n')
            if Match[0][0]:
                Atom = lambda x: self._Atom(int(x[0]),x[1],float(x[2]),float(x[3]))
            else:
                Atom = lambda x: self._Atom(int(x[0]),x[1],float(x[2]),spin=None)
            
            self.mulliken = [Atom(line.strip().split())
                             for line in lines if line.strip()]

@RegisterLinkJob
class Link716(LinkJob):
    """
    Representation and parser for the output of l716.exe. Displays
    information about Forces and Frequency Calculations.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l716.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    dipole : list
    units : list
    zeropoint : list
    thermal_energy : list
    enthalpy : list
    gibbs : list
    EContribs : list
        List of namedtuples with fields Name,Thermal,CV,S .
    IRSpectrum : str
        Text that corresponds to the IRSpectrum
    mode : str
        Either 'Forces', 'Freq' or 'Other'.
    frequencies : list
    freq_displacements : list
    """
    __slots__ = ('dipole', 'units','frequencies','EContribs', 'IRSpectrum',
                  'zeropoint','thermal_energy', 'enthalpy', 'gibbs', 'mode',
                  'freq_displacements')

    _token = 716

    re_Thermo = r'(Zero-point\scorrection=[\s\S]*)(?:\n\s*E\s\(Thermal\))'
    re_Thermo = re.compile(re_Thermo)
    re_EContrib =  r'(?:.*E\s\(Thermal\)\s*CV\s*S.*\n)' # Header
    re_EContrib += r'([\s\S]*\n)'                    # Body
    re_EContrib += r'(?:.*Q.*Log10\(Q\).*Ln\(Q\).*)' # Header of the next Table
    re_EContrib = re.compile(re_EContrib)
    re_IRSpectrum = re.compile(r'(.*IR\sSpectrum[\s\S]*?\n)(?:\s?\-\-+)')
    re_Frequencies =  r'(?:.*Frequencies\s\-\-)(.*)\n'
    re_Frequencies += r'(?:.*Red\.\smasses\s\-\-)(.*)\n'
    re_Frequencies += r'(?:.*Frc\sconsts\s*\-\-)(.*)\n'
    re_Frequencies += r'(?:.*IR\sInten\s*\-\-)(.*)'
    re_Frequencies = re.compile(re_Frequencies)
    re_freq_text = re.compile(r'  Atom[\s\S]*\n\n')
    re_freq_displacements = r'^ {3,5}'
    re_freq_displacements += r'[0-9]{1,3}\s*'*2
    re_freq_displacements += r'-?[0-9]*\.[0-9]*\s*'*3
    re_freq_displacements += r'[^\n]*\n'
    re_freq_displacements = re.compile(re_freq_displacements,re.MULTILINE)
    re_dipole = r'(?:Dipole.*\=)'+r'(.?[0-9]\.[0-9]*D[\-|\+][0-9]{2,})'*3
    re_dipole = re.compile(re_dipole)

    _Frequency = namedtuple('Frequency','freq redmass forcek IRint')
    _Displacement = namedtuple('FreqDisplacements','atomids atoms xyz')
    _EContrib = namedtuple('EContrib','Name Thermal CV S')

    def __init__(self,text,asEmpty=False):
        self.dipole = []
        self.units = [] # The first one corresponds to the non tabulated
        self.zeropoint = []
        self.thermal_energy = []
        self.enthalpy = []
        self.gibbs = []
        self.EContribs = []
        self.frequencies = []
        self.freq_displacements = []
        self.IRSpectrum = ''
        self.mode = ''
        if asEmpty:
            super().__init__('',716)
        else:
            super().__init__(text,716)
            if 'Forces' in self.text:
                self.mode = 'Forces'
            else:
                self.mode = 'Other'
            self._locate_dipole()
            self._locate_thermochemistry()
            self._locate_IRSpectrum()
            self._locate_frequencies()
            self._locate_freq_displacements()

    @Populates('dipole')
    @SilentFail
    def _locate_dipole(self):
        """
        Uses regex expressions compiled as class attributes to find the Dipole
        Vector.
        """
        cls = self.__class__
        Match = cls.re_dipole.findall(self.text)
        if Match:
            self.dipole = [float(i.replace('D','E')) for i in Match[0]]

    @Populates('frequencies')
    @SilentFail
    def _locate_frequencies(self):
        """
        Uses regex expressions compiled as class attributes to find the
        frequencies table.
        """
        cls = self.__class__
        Frequency = cls._Frequency
        Match = cls.re_Frequencies.findall(self.text)
        frequencies = []
        if Match:
            for i in Match:
                Aux = []
                for line in i:
                    Aux.append(tuple(map(float,line.strip().split())))
                for items in zip(*Aux):
                    frequencies.append(Frequency(*items))
            self.frequencies = frequencies
            self.mode = 'Freq'

    @Populates('freq_displacements')
    @SilentFail
    def _locate_freq_displacements(self): 
        """
        Uses regex expressions compiled as class attributes to find the
        cartesian displacements of the frequencies.
        """
        cls = self.__class__
        Displacement = cls._Displacement
        match = cls.re_freq_text.findall(self.text)
        if match:
            displacements = []
            for text in match[0].split('Atom'):
                if not text.strip(): 
                    continue
                lines = cls.re_freq_displacements.findall(text)
                atomlines = []
                for line in lines:
                    atomline = []
                    if not line.strip():
                        continue
                    atid,atnum,xyz = line.strip().split(maxsplit=2)
                    atomline.append(int(atid))
                    atomline.append(int(atnum))
                    xyz = [float(i) for i in xyz.split()]
                    for x,y,z in zip(xyz[0::3],xyz[1::3],xyz[2::3]): 
                        atomline.append((x,y,z))
                    atomlines.append(atomline)
                atomids,atomnums,*matrices = zip(*atomlines)
                for matrix in matrices: 
                    displacements.append(Displacement(atomids,atomnums,matrix))
            self.freq_displacements = displacements

    @Populates('zeropoint','thermal_energy','enthalpy','gibbs')
    @SilentFail
    def _locate_thermochemistry(self):
        """
        Uses regex expressions compiled as class attributes to find the
        thermochemistry computed properties.
        """
        cls = self.__class__
        EContrib = cls._EContrib
        Match = cls.re_Thermo.findall(self.text)
        if Match:
            lines = [line.strip() for line in Match[0].split('\n') if line.strip()]
            Properties = cycle([self.zeropoint,
                                self.thermal_energy,
                                self.enthalpy,
                                self.gibbs])
            for line,Property in zip(lines,Properties):
                Aux = line.split("=")[-1].strip().split()
                if len(Aux) == 2:
                    self.units.append(Aux[-1][1:-1])
                Property.append(float(Aux[0]))
        Match = cls.re_EContrib.findall(self.text)
        if Match:
            lines = [line.strip() for line in Match[0].split('\n') if line.strip()]
            Header = lines.pop(0)
            self.units.extend(Header.strip().split())
            for line in lines:
                Name, E, CV, S = line.strip().rsplit(maxsplit=3)
                item = EContrib(Name,float(E),float(CV),float(S))
                self.EContribs.append(item)

    @Populates('IRSpectrum')
    @SilentFail
    def _locate_IRSpectrum(self):
        """
        Uses regex expressions compiled as class attributes to find the IR
        spectrum.
        """
        cls = self.__class__
        Match = cls.re_IRSpectrum.findall(self.text)
        if Match:
            self.IRSpectrum = Match[0]

@RegisterLinkJob
class Link804(LinkJob):
    """
    Representation and parser for the output of l804.exe. Holds the
    information about the MP2 energy calculation

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l804.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    MP2 : float
        Potential energy with MP2 method
    SpinComponents : list
        List of namedtuples with the fields Name, T, E.
    """
    __slots__ = ('MP2', 'SpinComponents')

    _token = 804

    re_MP2 = re.compile(r'EUMP2\s=\s*(\-?[0-9]+\.[0-9]*D[\-|\+][0-9]{2,3})')

    _SpinComponent = namedtuple('SpinComponent','Name T E')

    def __init__(self,text,asEmpty=False):
        self.MP2 = None
        self.SpinComponents = [] # The first one corresponds to the non tabulated
        if asEmpty:
            super().__init__('',804)
        else:
            super().__init__(text,804)
            self._locate_SpinComponents()
            self._locate_MP2()

    def _locate_MP2(self):
        """ Looks for the keyword 'EUMP2=' and reads the MP2 Potential Energy """
        cls = self.__class__
        Match = cls.re_MP2.findall(self.text)
        if Match:
            self.MP2 = float(Match[0].replace('D','E'))
    def _locate_SpinComponents(self):
        """ Looks for the keywords 'Spin Components' and reads
        the Energies of the spin Components """
        Iterator = self._text_iterbyline()
        cls = self.__class__
        SpinComponent = cls._SpinComponent
        for line in Iterator:
            if 'Spin components' in line:
                break
        else: # fail silently
            return
        for _ in range(3):
            line = next(Iterator)
            # Standarize number of columns
            line = line.replace('=',' ').replace('E2',' ').replace('T2',' ')
            # Change Number Notation
            line = line.replace('D','E')
            # Store spin Component
            Name,T,E = line.split()
            self.SpinComponents.append(SpinComponent(Name,float(T),float(E)))
    def get_SCScorr(self):
        """ Calculates and returns the MP2(SCS) potential energy"""
        aa = self.SpinComponents[0].E
        ab = self.SpinComponents[1].E
        bb = self.SpinComponents[2].E
        return (aa+bb)/3.0 +(6.0*ab)/5.0

@RegisterLinkJob
class Link913(LinkJob):
    """Representation and parser for the output of l913.exe. Holds information
    about CCSD(T) calculations including MP4 Energy.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l913.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    MP4 : float
        MP4 potential energy
    CCSDT : float
        CCSD(T) potential energy
    """
    # Current Implementation only stores the MP4(SDQ) Energy, other energies
    # Should be implemented using a different name or reestructure all code

    __slots__ = ('MP4', 'CCSDT')

    _token = 913

    re_MP4 = r'[A-Z]*4\(S?DQ\)=\s*\-?[0-9]+\.[0-9]*D[\+|\-][0-9]{2,3}'
    re_MP4 = re.compile(re_MP4)
    re_ccsdt = re.compile(r'CCSD\(T\)=\s?(\-?[0-9]+\.[0-9]*D[\+|\-][0-9]{2,3})')

    def __init__(self,text,asEmpty=False):
        self.MP4 = None
        self.CCSDT = None
        if asEmpty:
            super().__init__('',913)
        else:
            super().__init__(text,913)
            self._locate_MP4()
            self._locate_CCSDT()
    def _locate_MP4(self):
        """ Looks for the First iteration and then looks for the keyword
        'UMP4(SDQ)' and reads the MP4 Potential Energy """
        cls = self.__class__
        Match = cls.re_MP4.findall(self.text)
        if Match and Match[-1].startswith('UMP4(SDQ)'):
            self.MP4 = float(Match[-1].split("=")[1].replace('D','E'))
    def _locate_CCSDT(self):
        """ Looks for the keywords 'Time for triples' which appears after the
        calculation has converged and reads the CCSD(T) Potential Energy """
        cls = self.__class__
        Match = cls.re_ccsdt.findall(self.text)
        if Match:
            self.CCSDT = float(Match[0].replace('D','E'))

@RegisterLinkJob
class Link914(LinkJob):
    """
    Representation and parser for the output of l914.exe. Holds information
    about TDDFT calculations excited states.

    Parameters
    ----------
    text : str
        text that corresponds to the output of the l913.exe
    asEmpty : bool
        Flag to not parse and store the information of the text.
        (defaults to False)

    Attributes
    ----------
    excitedstates : list
        list of excited states
    """

    __slots__ = ('excitedstates',)

    _token = 914

    _ExcitedState = namedtuple('ExcitedState',
                            'number energy wavelen OStrenght s2 transitions')
    _TransitionData = namedtuple('TransitionData',
                            'donor acceptor contribution isreversed')

    re_ExcitedState =  r'Excited\s*State\s*([0-9]{1,3})\:' # Excited State number
    re_ExcitedState += r'\s*(\S*)'     # Match Pointgroup
    re_ExcitedState += r'\s*(\S*)\seV' # Match Energy
    re_ExcitedState += r'\s*(\S*)\snm' # Match wavelength
    re_ExcitedState += r'\s*f\=(\S*)'  # Match Oscilator Strength
    re_ExcitedState += r'\s*\S*\=(\S*).*' # Match <S**2>
    re_ExcitedState = re.compile(re_ExcitedState)

    re_transition  = r'([0-9]{1,4}[A|B]?)\s*' # match first transition id
    re_transition += r'(\-\>|\<\-)\s*'        # Match "donation" direction
    re_transition += r'([0-9]{1,4}[A|B]?)\s*' # match second transition id
    re_transition += r'(\-?[0-9]*\.[0-9]*)'   # Match contribution
    re_transition = re.compile(re_transition)

    def __init__(self,text,asEmpty=False):
        self.excitedstates = []
        if asEmpty:
            super().__init__('',914)
        else:
            super().__init__(text,914)
            self._locate_ExcitedStates()

    @Populates('excitedstates')
    @SilentFail
    def _locate_ExcitedStates(self):
        """
        Does the logic of finding and slicing the text of the link into the
        different excited states
        """
        ExcitedState = self._ExcitedState
        excitedstates = []
        matches = list(self.re_ExcitedState.finditer(self.text))
        # Fail silently
        if not matches:
            return
        for current,new in zip(matches,matches[1:]):
            start,stop = current.end(),new.start()
            text = self.text[start:stop]
            number,_,energy,wavelen,OStrenght,s2 = current.groups()
            transitions = self._extract_transitions(text)
            excitedstates.append(ExcitedState(int(number),
                                              float(energy),
                                              float(wavelen),
                                              float(OStrenght),
                                              float(s2),
                                              transitions))
        self.excitedstates = excitedstates

    def _extract_transitions(self,text):
        """
        Given the slice of output text that corresponds to the transition
        contributions of a ExcitedState, returns each contribution as a
        TransitionData object.

        Parameters
        ----------
        text : str
            The slice of text that corresponds to the transition
            contributions of an ExcitedState.

        Returns
        -------
        transitions
            list of TransitionData instances.

        """
        TransitionData = self._TransitionData
        transitions_matches = self.re_transition.findall(text)
        transitions = []
        if transitions_matches:
            for donor,direction,acceptor,contribution in transitions_matches:
                isreversed = False
                if direction == '<-':
                    isreversed = True
                    donor,acceptor = acceptor,donor
                transition = TransitionData(donor,
                                            acceptor,
                                            float(contribution),
                                            isreversed)
                transitions.append(transition)
        return transitions

    def print_excitedstates(self,*ESnumbers,show_transitions=False):
        """
        Prints in the console the parameters in of the excited states selected 
        in ascending ordinal order.

        Parameters
        ----------

        *ESnumbers : int
            An undefined number of integrers. Only the excited states that are
            within the provided set will be displayed.
        show_transitions : bool, optional
            Allows control over the display of the transitions, by 
            default False
        """
        for ES in self.excitedstates:
            if ES.number not in ESnumbers:
                continue
            number,energy,wavelen,OStrenght,s2,_ = ES
            print(f'ExcitedState(number={number} energy={energy} ' \
                  f'wavelen={wavelen} OStrenght={OStrenght} ' \
                  f's2={s2} transitions=[...])')
            if show_transitions:
                for transition in ES.transitions:
                    print(f'\t{transition.donor} -> {transition.acceptor}'\
                          f'\t {transition.contribution}')

#@RegisterLinkJob
class Link9999(LinkJob):
    """
    parser for the output of l9999.exe. This Link has 2 termination modes:
    - normal, if so it will display a resume string
    - error, if so it will display or not information about the error
    Depending on the InternalJob, it might display a final structure or not,
    (before printing the resume string). In development.
    """
    _token = 9999
